# -*- coding: utf-8 -*-
"""
aid_start.py
some functions that I want to keep separate to make the main script a bit shorter
---------
@author: maikherbig
"""

import os,shutil,json

# =============================================================================
# def banner():
#     #generated using: https://www.ascii-art-generator.org/    
#     text = """
#           _____ _____  
#     /\   |_   _|  __ \ 
#    /  \    | | | |  | |
#   / /\ \   | | | |  | |
#  / ____ \ _| |_| |__| |
# /_/    \_\_____|_____/ 
# is starting...
# """
#     print(text)    
# =============================================================================
    


def get_default_dict(dir_settings):
    with open(dir_settings) as f:
        Default_dict = json.load(f)
        #Older versions of AIDeveloper might not have the Icon theme option->add it!
        if "Icon theme" not in Default_dict.keys():
            Default_dict["Icon theme"] = "Icon theme 1"
        if "Path of last model" not in Default_dict.keys():
            Default_dict["Path of last model"] = 'c:'+os.sep
        #Older versions of AIDeveloper might not have the Contrast, etc. Add it!
        if "Contrast On" not in Default_dict.keys():
            Default_dict["Contrast On"] = True
            Default_dict["Contrast min"] = 0.7
            Default_dict["Contrast max"] = 1.3
        if "Saturation On" not in Default_dict.keys():
            Default_dict["Saturation On"] = False
            Default_dict["Saturation min"] = 0.7
            Default_dict["Saturation max"] = 1.3
        if "Hue On" not in Default_dict.keys():
            Default_dict["Hue On"] = False
            Default_dict["Hue range"] = 0.08
        if "AvgBlur On" not in Default_dict.keys():
            Default_dict["AvgBlur On"] = True
            Default_dict["AvgBlur min"] = 0
            Default_dict["AvgBlur max"] = 5
        if "GaussBlur On" not in Default_dict.keys():
            Default_dict["GaussBlur On"] = False
            Default_dict["GaussBlur min"] = 0
            Default_dict["GaussBlur max"] = 5
        if "MotionBlur On" not in Default_dict.keys():
            Default_dict["MotionBlur On"] = False
            Default_dict["MotionBlur Kernel"] = "0,5"
            Default_dict["MotionBlur Angle"] = "-10,10"
        if "Image_import_dimension" not in Default_dict.keys():
            Default_dict["Image_import_dimension"] = 150
            Default_dict["Image_import_interpol_method"] = "Lanczos"

        if "spinBox_batchSize" not in Default_dict.keys():
            Default_dict["spinBox_batchSize"] = 32
        if "doubleSpinBox_learningRate_Adam" not in Default_dict.keys():
            Default_dict["doubleSpinBox_learningRate_Adam"] = 0.001
        if "doubleSpinBox_learningRate_SGD" not in Default_dict.keys():
            Default_dict["doubleSpinBox_learningRate_SGD"] = 0.01
        if "doubleSpinBox_learningRate_RMSprop" not in Default_dict.keys():
            Default_dict["doubleSpinBox_learningRate_RMSprop"] = 0.001
        if "doubleSpinBox_learningRate_Adagrad" not in Default_dict.keys():
            Default_dict["doubleSpinBox_learningRate_Adagrad"] = 0.01
        if "doubleSpinBox_learningRate_Adadelta" not in Default_dict.keys():
            Default_dict["doubleSpinBox_learningRate_Adadelta"] = 1.0
        if "doubleSpinBox_learningRate_Adamax" not in Default_dict.keys():
            Default_dict["doubleSpinBox_learningRate_Adamax"] = 0.002
        if "doubleSpinBox_learningRate_Nadam" not in Default_dict.keys():
            Default_dict["doubleSpinBox_learningRate_Nadam"] = 0.002




    return Default_dict

def save_default_dict(dir_settings):
    #In case the standard dictionary is overwritten, this function can create a
    #initial disctionary that works
    #Users can open the .json with a text editor and change stuff if needed
    Default_dict = {"norm_methods":["None","Div. by 255","StdScaling using mean and std of each image individually","StdScaling using mean and std of all training data"],\
    "Input image size":32,"Normalization":"Div. by 255",\
    "Nr. epochs":2500,"Keras refresh after nr. epochs": 2,\
    
    "Image_import_dimension":150,"Image_import_interpol_method":"Lanczos",\
    
    "Horz. flip":False,"Vert. flip":True,"rotation":3,\
    "width_shift":0.001,"height_shift":0.001,"zoom":0.001,"shear":0.005,\
    
    "Brightness refresh after nr. epochs": 1,\
    "Brightness add. lower":-15,"Brightness add. upper":15,\
    "Brightness mult. lower":0.7,"Brightness mult. upper":1.3,\
    
    "Gaussnoise Mean":0,"Gaussnoise Scale":3.0,\
    
    "Contrast On": True, "Contrast min":0.7, "Contrast max":1.3,\
    "Saturation On": False, "Saturation min":0.7, "Saturation max":1.3,\
    "Hue On": False, "Hue range":0.08,\
    
    "AvgBlur On": True, "AvgBlur min":0, "AvgBlur max":5,\
    "GaussBlur On": False, "GaussBlur min":0, "GaussBlur max":5,\
    "MotionBlur On": False, "MotionBlur Kernel":"0,5", "MotionBlur Angle":"-10,10",\
    
    "Layout": "Normal", "Icon theme":"Icon theme 1",\
    "Path of last model":"c:\\",\

    "spinBox_batchSize":128,\
    "doubleSpinBox_learningRate_Adam":0.001,\
    "doubleSpinBox_learningRate_SGD":0.01,\
    "doubleSpinBox_learningRate_RMSprop":0.001,\
    "doubleSpinBox_learningRate_Adagrad":0.01,\
    "doubleSpinBox_learningRate_Adadelta":1.0,\
    "doubleSpinBox_learningRate_Adamax":0.002,\
    "doubleSpinBox_learningRate_Nadam":0.002}

    with open(dir_settings, 'w') as f:
        json.dump(Default_dict,f)




def keras_json_replace(keras_json_path,json_exists=True):
    if json_exists:
        #Inform user!
        print("I found a keras.json file in your home-directory which has options AID does not accept.\
              This file will be copied to keras_beforeAID_x.json and a new keras.json\
              is written with valid options for AID.")
        i=0
        while os.path.isfile(os.path.expanduser('~')+os.sep+'.keras'+os.sep+'keras_beforeAID_'+str(i)+'.json'):
            i+=1
        shutil.copy(keras_json_path, os.path.expanduser('~')+os.sep+'.keras'+os.sep+'keras_beforeAID_'+str(i)+'.json')
    
    #Write new keras.json:
    with open(os.path.expanduser('~')+os.sep+'.keras'+os.sep+'keras.json','w') as f:
        new_settings = """{\n    "image_dim_ordering": "tf", \n    "epsilon": 1e-07, \n    "floatx": "float32", \n    "backend": "tensorflow"\n}"""                       
        f.write(new_settings)

def keras_json_check(keras_json_path):
    if os.path.isfile(keras_json_path):
        with open(keras_json_path, 'r') as keras_json:
            keras_json=keras_json.read()
        keras_json = json.loads(keras_json)
        keys_keras_json = keras_json.keys() #contained keys 
        keys_expected = ['image_dim_ordering','backend','epsilon','floatx'] #expected keys
        #are those keys present in keys_keras_json?
        keys_present = [key in keys_expected for key in keys_keras_json]
        keys_present = all(keys_present) #are all true?
        checks = []
        if keys_present==True:#all keys are there, now check if the value is correct
            checks.append( keras_json['image_dim_ordering']=="tf" )   
            checks.append( keras_json['backend']=="tensorflow" )    
            checks.append( keras_json['epsilon']==1e-07 )   
            checks.append( keras_json['floatx']=="float32" )    
            checks = all(checks) #are all true?
            if checks==True:
                #keras.json is fine, no need to overwrite it
                return
            else:#some values are different. AID need to write new keras.json
                keras_json_replace(keras_json_path)
        
        else:#some keys are missing
            keras_json_replace(keras_json_path)
    
    else:#there exists NO keras.json! Very likely the user opened AID for the first time :)
        #Welcome the user
        print("A warm welcome to your first session in AID :)\
             \nIn case of any issues just write me a mail: maik.herbig@tu-dresden.de\
             \nor use the Issues section on the github page of AID:\
             \nhttps://GitHub.com/maikherbig/AIDeveloper\
             ")
        keras_json_replace(keras_json_path,False)

def get_tooltips():
    tooltips = {}
    tooltips["groupBox_dragdrop"] = "Drag and drop files (.rtdc) or folders with images here. Valid .rtdc files have to contain at least 'images', 'pos_x' and 'pos_y'. If folders with images are dropped, the contents will be converted to a single .rtdc file (speeds up loading in the future).<br>After dropping data, you can specify the ‘Class’ of the images and if it should be used for training (T) or validation (V).<br>Double-click on the filename to show a random image of the data set. The original and cropped (using 'Input image size') image is shown<br>Click on the button 'Plot' to open a popup where you can get a histogram or scatterplot of enclosed data.<br>'Cells/Epoch' defines the nr. of images that are used in each epoch for training. Random images are drawn in each epoch during training. For the validation set, images are drawn once at the beginning and kept constant.<br>Deactivate 'Shuffle' if you don't want to use random data. Then all images of this file are used.<br>Zoom allows you to increase or decrease resolution. Zoom=1 does nothing; Zoom=2 zooms in; Zoom=0.5 zooms out. This is useful if you have data acquired at 40x but you want to use it to train a model for 20x data. Use 'Options'->'Zooming order' to define the method used for zooming.<br>Hint for RT-DC users: Use ShapeOut to gate for particular subpopulations and save the filtered data as .rtdc. Make sure to export at least 'images', 'pos_x' and 'pos_y'."
    tooltips["groupBox_DataOverview"] = "The numbers of events of each class are added up. To do so the properties of each file are read. This happens each time you click into the table above. Unfortunately, reading is quite slow, so maybe disable this box, while you are assembling your data-set (especially when working with many large files). Use the column 'Name' to specify custom class-names, which help you later to remember, the meaning of each class. This table is saved to meta.xlsx when training is started."
    tooltips["tab_ExampleImgs"] = "Show random example images of the training data"
    tooltips["comboBox_ModelSelection"] = "Select the model architecture. MLP means Multilayer perceptron. Currently only MLPs are fast enough for AI based sorting. The definition of the architectures can be found in the model_zoo.py. If you want to implement custom Neural net architectures, you can edit model_zoo.py accordingly. Restart AIDeveloper in order to re-load model_zoo and import the new definitions"
    tooltips["radioButton_NewModel"] = "Select a model architecture in the dropdown menu"
    tooltips["radioButton_LoadRestartModel"] = "Load an existing architecture (from .model or .arch), and start training using random initial weights"
    tooltips["radioButton_LoadContinueModel"] = "Load an existing model (.model) and continue fitting. This option could be used to load a trained model to optimize it (transfer learning)"
    tooltips["lineEdit_LoadModelPath"] = "When you use 'Load and restart' or 'Load and continue', the filename of the chosen model will apprear here"
    tooltips["label_Crop"] = "Models need a defined input size image. Choose wisely since cutting off from large objects might result in information loss."
    tooltips["label_Normalization"] = "Image normalization method.<br><br>None: No normalization is applied.<br><br>'Div. by 255': Each input image is divided by 255 (useful since pixel values go from 0 to 255, so the result will be in range 0-1).<br><br>StdScaling using mean and std of each image individually: The mean and standard deviation of each input image itself is used to scale it by first subtracting the mean and then dividing by the standard deviation.<br><br>StdScaling using mean and std of all training data: First, all pixels of the entire training dataset are used to calc. a mean and a std. deviation. This mean and standard deviation is used to scale images during training by first subtracting the mean and then dividing by the standard deviation.<br><br>Only 'None' and 'Div. by 255' are currently supported in the Sorting-Software"
    tooltips["label_colorMode"] = "The Color Mode defines the input image depth. Color images have three channels -RGB. Grayscale images only have a single channel. Models are automatically built accordingly. Models trained for Grayscale cannot be applied to RGB images (unless images are converted). Same is true the other way around."
    tooltips["label_nrEpochs"] = "Number of epochs (iterations over the training data). In each epoch, a subset of the training data is used to update the weights if 'Shuffle' is on."
    tooltips["label_paddingMode"] = "Padding mode to use when requested image size is larger than available images.<br>constant: Pads a constant value (0).<br>edge: Pads with the edge values of array.<br>reflect: Pads with the reflection of the vector mirrored on the first and last values of the vector along each axis.<br>symmetric: Pads with the reflection of the vector mirrored along the edge of the array.<br>wrap: Pads with the wrap of the vector along the axis. The first values are used to pad the end and the end values are used to pad the beginning.<br>Delete: dont do padding, but delete images, if the cell is too close to the border.<br>Text copied from https://docs.scipy.org/doc/numpy/reference/generated/numpy.pad.html"

    tooltips["pushButton_modelname"] = "Define path and filename for the model you want to fit."

    tooltips["label_zoomIcon"] = "Define which zoom order should be used. A zoom order of 0 would mean nearest neighbor interploation. Zooming is applied when data is loaded into AID. Zooming is only applied of the Zoom factor given in the table above is not =1."

    tooltips["radioButton_cpu"] = "Train model on CPU."
    tooltips["radioButton_gpu"] = "Train model on GPU."
    tooltips["comboBox_gpu"] = "Select the device your model should be trained on. 'Multi-GPU' will train the model on multiple GPUs in parallel."

    
    tooltips["label_memory"] = "Limit the amount of memory used. 1.0 means 100%; 0.5 means 50%; Typically, 0.7 is a good value for a system with one GPU."
    tooltips["label_RefreshAfterEpochs"] = "Affine image augmentation takes quite long; so maybe use the same images for this nr. of epochs"
    tooltips["tab_kerasAug"] = "Define settings for  affine image augmentations"
    tooltips["checkBox_HorizFlip"] = "Flip some training images randomly along horiz. axis (left becomes right; right becomes left)"
    tooltips["checkBox_VertFlip"] = "Flip some training images randomly along vert. axis (bottom up; top down)"
    tooltips["label_Rotation"] = "Degree range for random rotations"
    tooltips["label_width_shift"] = "Define random shift of width<br>Fraction of total width, if &lt; 1. Otherwise pixels if>=1.<br>Value defines an interval (-width_shift_range, +width_shift_range) from which random numbers are created."
    tooltips["label_height_shift"] = "Define random shift of height<br>Fraction of total height if &lt; 1. Otherwise pixels if>=1.<br>Value defines an interval (-height_shift_range, +height_shift_range) from which random numbers are created."
    tooltips["label_zoom"] = "Range for random zoom"
    tooltips["label_shear"] = "Shear Intensity (Shear angle in counter-clockwise direction in degrees)"
    tooltips["spinBox_RefreshAfterEpochs"] = "Affine image augmentation takes quite long; so maybe use the same images for this nr. of epochs"
    tooltips["label_RefreshAfterNrEpochs"] = "Brightness augmentation is fast, so you can probably refresh images for each epoch (set to 1)"
    tooltips["groupBox_BrightnessAugmentation"] = "Define add/multiply offset to make image randomly slightly brighter or darker. Additive offset (A) is one number that is added to all pixels values; Multipl. offset (M) is a value to multiply each pixel value with: NewImage = A + M*Image"
    tooltips["label_Plus"] = "Brightness augmentation by adding a random value from given range."
    tooltips["spinBox_PlusLower"] = "Define lower threshold for additive offset"
    tooltips["spinBox_PlusUpper"] = "Define upper threshold for additive offset"
    tooltips["label_Mult"] = "Brightness augmentation by multiplying the pixel values of the image with a given value (random value from given range)."

    tooltips["doubleSpinBox_MultLower"] = "Define lower threshold for multiplicative offset"
    tooltips["doubleSpinBox_MultUpper"] = "Define upper threshold for multiplicative offset"
    tooltips["groupBox_GaussianNoise"] = "Define Gaussian Noise, which is added to the image"
    tooltips["label_GaussianNoiseMean"] = "Define the mean of the Gaussian noise. Typically this should be zero. If you use a positive number it would mean that your noise tends to be positive, i.e. bright."
    tooltips["label_GaussianNoiseScale"] = "Define the standard deviation of the Gaussian noise. A larger number means a wider distribution of the noise, which results in an image that looks more noisy. Prefer to change this parameter over chainging the mean."
    tooltips["groupBox_colorAugmentation"] = "Define methods to randomly alter the contrast (applicable for grayscale and RGB), saturation (RGB only) or hue (RGB only) of your images."
    tooltips["checkBox_contrast"] = "Augment contrast using a random factor. Applicable for Grayscale and RGB. Left spinbox (lower factor) has to be >0. '0.70' to '1.3' means plus/minus 30% contrast (at random)"
    tooltips["checkBox_saturation"] = "Augment saturation using a random factor. Applicable for RGB only. Left spinbox (lower factor) has to be >0. '0.70' to '1.3' means plus/minus 30% saturation (at random)"
    tooltips["checkBox_hue"] = "Augment hue using a random factor. Applicable for RGB only. Left spinbox (lower factor) has to be >0. '0.70' to '1.3' means plus/minus 30% hue (at random)"
    tooltips["groupBox_blurringAug"] = "Define methods to randomly blur images."
    tooltips["label_motionBlurKernel"] = "Define kernels by giving a range [min,max]. Values in this range are then randomly picked for each image"
    tooltips["lineEdit_motionBlurAngle"] = "Define angle for the motion blur by defining a range \"min degree,max degree\". Values in this range are then randomly picked for each image"
    tooltips["label_avgBlurMin"] = "Define the minimum kernel size for average blur"
    tooltips["spinBox_gaussBlurMax"] = "Define the maximum kernel size for gaussian blur"
    tooltips["checkBox_motionBlur"] = "Apply random motion blurring. Motion blur is defined by an intensity ('kernel') and a direction ('angle'). Please define a range for 'kernel' and 'angle'. AID will pick a random value (within each range) for each image."
    tooltips["spinBox_avgBlurMin"] =  "Define the minimum kernel size for average blur"
    tooltips["spinBox_gaussBlurMin"] = "Define the minimum kernel size for gaussian blur"
    tooltips["label_motionBlurAngle"] = "Define a range of angles for the motion blur: 'min degree,max degree'. Values from this range are then randomly picked for each image"
    tooltips["label_gaussBlurMin"] = "Define the minimum kernel size for gaussian blur"
    tooltips["checkBox_gaussBlur"] = "Apply random gaussian blurring. For gaussian blurring, a gaussian kernel of defined size is used. Please define a min. and max. kernel size. For each image a random value is picked from this range to generate a gaussian kernel."
    tooltips["spinBox_avgBlurMax"] = "Define the maximum kernel size for average blur"
    tooltips["label_gaussBlurMax"] = "Define the maximum kernel size for gaussian blur"
    tooltips["label_gaussBlurMax"] = "Define the maximum kernel size for gaussian blur"
    tooltips["checkBox_avgBlur"] = "Apply random average blurring. Define a range of kernel sizes for the average blur (min. and max. kernel size). Values from this range are then randomly picked for each image. To blur an image, all pixels within the kernel area used to compute an average pixel value. The central element of the kernel area in the image is then set to this value. This operation is carried out over the entire image."
    tooltips["label_avgBlurMax"] = "Define the maximum kernel size for average blur"
    tooltips["spinBox_avgBlurMin"] = "Define the minimum kernel size for average blur"
    tooltips["spinBox_avgBlurMax"] = "Define the maximum kernel size for average blur"
    tooltips["lineEdit_motionBlurKernel"] = "Define kernels by giving a range \"min,max\". Values from this range are then randomly picked for each image"
    tooltips["groupBox_expertMode"] = "Expert mode allows changing the learning rate and you can even set parts of the neural net to \'not trainable\' in order to perform transfer learning and fine tune models. Also dropout rates can be adjusted. When expert mode is turned off again, the initial values are applied again."
    tooltips["groupBox_learningRate"] = "Change the learning rate. The default optimizer is \'adam\' with a learning rate of 0.001"
    tooltips["checkBox_trainLastNOnly"] = "When checked, only the last n layers of the model, which have >0 parameters will stay trainable. Layers before are set to trainable=False. Please specify n using the spinbox. After this change, the model has to be recompiled, which means the optimizer values are deleted."  
    tooltips["spinBox_trainLastNOnly"] = "Specify the number of last layer in your model that should be kept trainable. Only layers with >0 parameters are counted. Use the checkbox to apply this option. After this change, the model has to be recompiled, which means the optimizer values are deleted."
    tooltips["checkBox_trainDenseOnly"] = "When checked, only the dense layers are kept trainable (if they have >0 parameters). Other layers are set to trainable=False. After this change, the model has to be recompiled, which means the optimizer values are deleted."
    tooltips["label_batchSize"] = "Number of samples per gradient update. If unspecified, batch_size will default to 32. (Source: Keras documentation)"
    tooltips["label_epochs"] = "Number of epochs to train the model on an identical batch."    
 
    tooltips["groupBox_learningRate"] = "The learning rate defines how much model weights are changed in each training iteration."    
    tooltips["doubleSpinBox_learningRate"] = "Define a constant value for the learning rate. The learning rate defines how strong parameters are changed in each training iteration."    
    tooltips["radioButton_LrCycl"] = "Apply cyclical learning rate schedule. Here, the learning rate oszillates between two bounds (Min/Max). After each processed batch, the lr is adjusted. Button 'LR Screening' on main screen leads to functions that allows to find sensible bounds. Cyclical learning rate was orignally defined in the paper 'Cyclical Learning Rates for Training Neural Networks': https://arxiv.org/abs/1506.01186."    
    tooltips["label_cycLrMin"] = "Lower bound for cyclical learning rate."    
    tooltips["label_cycLrMax"] = "Upper bound for cyclical learning rate."    
    tooltips["comboBox_cycLrMethod"] = "Method for changing the learning rate."    
    tooltips["pushButton_cycLrPopup"] = "Advanced settings for cyclical learning rates."    
    
    tooltips["radioButton_LrExpo"] = "Apply exponentially decreasing learning rates (LR). LR is changed at every new batch. Therefore, the number of iterations (decay steps) in one epoch is: Nr_of_training_images/batch_size.<br>Equation: LR = initial_LR * decay_rate ^ (batch_iteration_nr / decay_steps)."    
    tooltips["label_expDecInitLr"] = "Initial learning rate for exponentially decreasing learning rates (LR).<br>Equation: LR = initial_LR * decay_rate ^ (batch_iteration / decay_steps)."    
    tooltips["label_expDecSteps"] = "Decay steps: Nr. of batch-updates it should take till LR decreases to decay_rate*inital_LR.<br>Equation: LR = initial_LR * decay_rate ^ (batch_iteration / decay_steps)."    
    tooltips["label_expDecRate"] = "Decay rate: choose value between 0 and 1. Lower means earlier drop of LR. 1 means never decrease learning rate.<br>Equation: LR = initial_LR * decay_rate ^ (batch_iteration / decay_steps)."    

    tooltips["checkBox_dropout"] = "If your model has one or more dropout layers, you can change the dropout rates here. Insert into the lineEdit one value (e.g. 0.5) to apply this one value to all dropout layers, or insert a multiple values to specify the dropout rates for each dropout layer individually (e.g. for three dropout layers insert: 0.2 , 0.5 , 0.25<br>. The model will be recompiled, but the optimizer weights are not deleted."
    tooltips["checkBox_partialTrainability"] = "Partial trainability allows you to make parts of a layer non-trainable. Hence, this option makes most sense in combination with 'Load and continue' training a model. After checking this box, the model you chose on 'Define model'-tab is initialized. The line on the right shows the trainability of each layer in the model. Use the button on the right ('...') to open a popup menu, which allows you to specify individual trainabilities for each layer."
    tooltips["label_expt_loss"] = "Specify which loss function should be used. From Keras documentation: 'The purpose of loss functions is to compute the quantity that a model should seek to minimize during training.'"
    tooltips["label_optimizer"] = "Specify which optimizer should be used."
    tooltips["groupBox_lossOptimizer"] = "The loss function defines how much error the neural net does in prediction. This error depends on the weights of the neural net and upon updating the weights the error should be decreased. Hence, the loss is seen as a function of the model weights. Typical function to compute loss is categorical_crossentropy. Optimizing the weights requires to perform updates of the model weights in a reasonable magnitude and the so called 'Optimizer' is responsible for this task."

    tooltips["checkBox_lossW"] = "Specify scalar coefficients to weight the loss contributions of different classes."
    tooltips["groupBox_expertMetrics"] = "Define metrics, that are computed after each training iteration ('epoch'). Those metrics are can then also be displayed in real-time during training and are tracked/saved in the meta.xlsx file. Each model where any metric for the validation set breaks a new record is saved (minimum val. loss achived -> model is saved. maximum val. accuracy achieved-> model is saved)."
    tooltips["checkBox_expertAccuracy"] = "Compute accuracy and validation accuracy after each epoch. Each model, where the corresponding metric for the validatio-set achieves a new record will be saved."
    tooltips["checkBox_expertF1"] = "Compute F1 and validation F1 score after each epoch. Each model, where the corresponding metric for the validatio-set achieves a new record will be saved."
    tooltips["checkBox_expertPrecision"] = "Compute precision and validation precision after each epoch for each class. Each model, where the corresponding metric for the validatio-set achieves a new record will be saved."
    tooltips["checkBox_expertRecall"] = "Compute recall and validation recall after each epoch for each class. Each model, where the corresponding metric for the validatio-set achieves a new record will be saved."
    tooltips["pushButton_FitModel"] = "Afer defining all model parameters, hit this button to build/initialize the model, load the data to RAM (if 'Load data to RAM' is chosen in 'Edit') and start the fitting process. You can also do only the initialization to check the model summary (appears in textbox on the left)."
    tooltips["pushButton_Live"] = "Load and display the history of the model which is currently fitted"
    tooltips["pushButton_LoadHistory"] = "Select a history file to be plotted"
    tooltips["lineEdit_LoadHistory"] = "Enter path/filename of a meta-file (meta.xlsx). The history is contained in this file."
    tooltips["tableWidget_HistoryItems"] = "Information of the history file is shown here<br>Double-click to adjust color"
    tooltips["checkBox_rollingMedian"] = "Check to add a rolling median. Use the slider to change the window size for which the median is computed."
    tooltips["horizontalSlider_rollmedi"] = "Use this slider to change the window size for the rolling median between 1 and 50."
    tooltips["checkBox_linearFit"] = "Check if you want to add a liner fit. A movable region will appear. Only data inside this region will be used for the fit."
    tooltips["pushButton_LoadModel"] = "Load a model from disk for conversion to other formats. Please specify the format of the model using the dropbox above. Next, define the target format using the dropdown menu on the right-> and finally, hit 'Convert'."
    tooltips["pushButton_convertModel"] = "Convert chosen model to the format indicated by the dropbox above. This might be useful to deply models to other platforms. AIDeveloper is only compatible with Keras TensorFlow models to perform inference. For usage of the model with soRT-DC, please convert it to .nnet (currently only MLPs are supported by soRT-DC software!)."
    tooltips["lineEdit_ModelSelection_2"] = "Model architecture name, read from meta.xlsx ('Chosen Model') is displayed here"
    tooltips["tableWidget_Info_2"] = "Specify validation data via 'Build'-tab or load .rtdc file (->From .rtdc).<br>Use column 'Name' to specify proper cell names (for presentation purposes).<br>Use column 'Clr' to specify the plotting color of that class"
    tooltips["lineEdit_LoadModel_2"] = "Enter path/filename of a model (.model)"
    tooltips["lineEdit_LoadModel_3"] = "Model architecture for (when using 'New' model) or path/filename of a model (.model) when using (load and continue/restart)"

    tooltips["pushButton_ExportValidToNpy"] = "Export the validation data (images and labels). Optionally the cropped images can exported->use 'Options'->'Export' to change. Normalization method of the chosen model is not yet applied. Please use the \'Build\' tab to define data"
    tooltips["pushButton_ImportValidFromNpy"] = "Import validation data (images from .rtdc and labels from .txt) file. Cropped and non-cropped images can be imported. If necessary they will be cropped to the correct format. If the loaded images are smaller than the required size, there will be zero-padding."
    tooltips["groupBox_InferenceTime"] = "Inference time is the time required to predict a single image. To get a meaningful value, several (define how many using spinbox->) images are predicted one by one. The given Nr. (spinbox->) is divided by 10. The resulting nr. of images is predicted one by one and an average computing time is obtained. This process is repqeated 10 times"
    tooltips["label_SortingIndex"] = "Specify the class you intend to sort for ('Target cell type'). This is important if you want to know the final concentration in the target region"
    tooltips["checkBox_SortingThresh"] = "Specify a probability threshold above which a cell is classified as a target cell (specified using 'Sorting class'). Typically cells with a probability above 0.5 are sorted, but you could also increase the threshold in order to only have cells in the target region that are more confidently classified"
    tooltips["comboBox_probability_histogram"] = "Select a plotting style for the probability plot<br>Style1: Only outline, width 5<br>Style2: Only outline, width 10<br>Style4: Filled hist, alpha 0.6<br>Style3: Filled hist, alpha 0.7<br>Style5: Filled hist, alpha 0.8"
    tooltips["groupBox_3rdPlot"] = "ROC (Receiver Operating Characteristic) curves summarize the trade-off between the true positive rate and false positive rate for a predictive model using different probability thresholds.<br>Precision-Recall curves summarize the trade-off between the true positive rate and the positive predictive value for a predictive model using different probability thresholds.<br>ROC curves are appropriate when the observations are balanced between each class, whereas precision-recall curves are appropriate for imbalanced datasets."
    tooltips["comboBox_3rdPlot"] = "Use the combobox to define what is shown in the third plot. Some options might only be valid for binary problems. For such cases please use the spinboxes Indx1 and Indx2 to define two cell types"
    tooltips["label_Indx1"] = "Some options of the combobox (\'Third plot\') are only valid for binary problems. For such cases please use the spinboxes Indx1 and Indx2 to define two cell types"
    tooltips["tableWidget_CM1"] = "Confusion matrix shows the total Nrs. of cells. Doubleclick a field in the matrix to save the corresponding images to .rtdc"
    tooltips["tableWidget_CM2"] = "Normalized Confusion matrix shows the relative amount of cells. Doubleclick a field in the matrix to save the corresponding images to .rtdc "
    text_scikit_learn = "Compute precision, recall, F-measure and support for each class<br>\
    The precision is the ratio tp / (tp + fp) where tp is the number of \
    true positives and fp the number of false positives. The precision is \
    intuitively the ability of the classifier not to label as positive a \
    sample that is negative.<br>The recall is the ratio tp / (tp + fn) where \
    tp is the number of true positives and fn the number of false negatives. \
    The recall is intuitively the ability of the classifier to find all the \
    positive samples.<br>The F-beta score can be interpreted as a weighted \
    harmonic mean of the precision and recall, where an F-beta score reaches \
    its best value at 1 and worst score at 0. The F-beta score weights recall \
    more than precision by a factor of beta. beta == 1.0 means recall and \
    precision are equally important.<br>The support is the number of occurrences \
    of each class in y_true.<br>micro average - averaging the total true \
    positives false negatives and false positives<br>macro average - averaging \
    the unweighted mean per label<br>weighted average - averaging the \
    support-weighted mean per label<br>(Source:scikit-learn.org)"
    tooltips["tableWidget_AccPrecSpec"] = "Classification Metrics appear after hitting 'Update Plots'<br>Copy to clipboad by clicking somewhere on table<br><br>"+text_scikit_learn+""
    tooltips["groupBox_probHistPlot"] = "Probability histogram appears after hitting 'Prob. hist'. It shows the predicted probability of each cell to remain to class 'Sorting clas'. Colors indicate the true label. A good model returns high probabilities for cells of 'Sorting class', while cells of other indices (different cell types) have very low probabilities. This plot also allows to guess a reasonable 'Sorting threshold'"
    tooltips["comboBox_chooseRtdcFile"] = "Choose a file"
    tooltips["comboBox_featurey"] = "Dropdown menu shows all availble features of the chosen .rtdc-file. The chosen feature will be plotted on the y-axis"
    tooltips["widget_histx"] = "Histogram projection of the x-dimension"
    tooltips["widget_histy"] = "Histogram projection of the y-dimension"   
    tooltips["horizontalSlider_cellInd"] = "Use this slider to choose a cell in the data-set. The respective cell and trace will be shown in tab 'Peakdetection'"  
    tooltips["spinBox_cellInd"] = "Use this to index and choose an event in the data-set. The respective cell and trace will be shown in tab 'Peakdetection'"
    tooltips["widget_scatter"] = "Click into this scatterplot to choose an event. The respective cell and trace will be shown in tab 'Peakdetection'"
    tooltips["checkBox_fl1"] = "Check this if you want to plot the trace for Fl1. Automatic peak finding (Highest x% in 'Peakdetection'-tab) will then also search in those traces."
    tooltips["checkBox_fl2"] = "Check this if you want to plot the trace for Fl2. Automatic peak finding (Highest x% in 'Peakdetection'-tab) will then also search in those traces."
    tooltips["checkBox_fl3"] = "Check this if you want to plot the trace for Fl3. Automatic peak finding (Highest x% in 'Peakdetection'-tab) will then also search in those traces."
    tooltips["checkBox_centroid"] = "Check this if you want to plot the centroid of a chosen event in 'Show cell' on 'Peakdetection'-tab)."
    tooltips["pushButton_selectPeakPos"] = "After you moved the range on the trace-plot you can select a peak using this button. This data of the peak will be shown in the table in the right and will be used to fit the peak-detection model."
    tooltips["pushButton_selectPeakRange"] = "After you changed the range on the trace-plot you can select the range-width using this button. This range will be shown in the table below and will be used in the peak-detection model."
    tooltips["pushButton_highestXPercent"] = "The highest x% of FLx_max peaks are looked up for each x (=1,2,3 if Checkboxes are activated) and inserted into the table on the right."
    tooltips["pushButton_removeSelectedPeaks"] = "Select a peak in the table or in the plot on the right and remove it using this button."    
    tooltips["pushButton_removeAllPeaks"] = "Remove all peaks from the table the right."
    tooltips["widget_showSelectedPeaks"] = "Scatterplot shows all selcted peaks. Click on a peak to highlight the respective position in the table. After that you can also use the button 'Selected' to remove this point"
    tooltips["tableWidget_showSelectedPeaks"] = "Table shows all seelcted peaks. After clicking on a row you can use the button 'Selected' to remove this point"
    tooltips["groupBox_showCell"] = "Plot shows image of the recorded cell and the respective fluorescence traces (optional- use checkboxes to show FL1/2/3). Optionally the centroid is shown."
    tooltips["pushButton_updateScatterPlot"] = "Hit this button to read the chosen features and plot the scatterplot above."
    tooltips["tableWidget_peakModelParameters"] = "After fitting a peak-detection model, the model parameters are shown here. Each parameter can also be manipulated right here."
    tooltips["comboBox_peakDetModel"] = "Choose a peak detection model."
    tooltips["pushButton_fitPeakDetModel"] = "Start fitting a model usig the selected peaks shown above"
    tooltips["pushButton_SavePeakDetModel"] = "Save model to an excel file"
    tooltips["pushButton_loadPeakDetModel"] = "Load peak detection model from an excel file"
    tooltips["radioButton_exportSelected"] = "Apply the peak detection model only on the single chosen file"
    tooltips["radioButton_exportAll"] = "Apply the peak detection model on all files on the 'Build'-Tab"
    tooltips["modelsaved_success"] = "The model was successfully saved. 'Load and continue' was automatically selected  in 'Define Model'-Tab, so your model with partial trainability will be loaded when you start fitting. The model architecture is documented in each .model file and the .arch file. Change of partial trainability during training is not supported yet (but it is theoretically no problem to implement it)."


    ###############Fitting popup window##################
    tooltips["checkBox_realTimePlotting_pop"] = "Plot model metrics like accuracy, val. accuracy... in real time. Please first hit 'Update plot' to initiate the plotting region."
    tooltips["label_realTimeEpochs_pop"] = "Define how many of the last epochs should be plotted in real time. 250 means the last 250 epochs are plotted"
    tooltips["pushButton_clearTextWindow_pop"] = "Clear the text window (fitting info)."
    tooltips["checkBox_ApplyNextEpoch"] = "Changes made in this window will be applied at the next epoch."
    tooltips["checkBox_saveEpoch_pop"] = "Save the model, when the current epoch is done"
    tooltips["pushButton_Pause_pop"] = "Pause fitting, push this button again to continue."
    tooltips["pushButton_Stop_pop"] = "Stop fitting entirely, Close this window manually, after the progressbar shows 100%."
    tooltips["actioncpu_merge"] = "Identify whether to force merging model weights under the scope of the CPU or not. Source: https://www.tensorflow.org/api_docs/python/tf/keras/utils/multi_gpu_model"
    tooltips["actioncpu_relocation"] = "Identify whether to create the model's weights under the scope of the CPU. If the model is not defined under any preceding device scope, you can still rescue it by activating this option. Source: https://www.tensorflow.org/api_docs/python/tf/keras/utils/multi_gpu_model"
    tooltips["actioncpu_weightmerge"] = "Uses tf.device('/cpu:0') prior constriction model to manage merging of weights using CPU."
    
    ############Confusion matrix show image popup
    tooltips["groupBox_model"] = "Properties of the model."
    tooltips["pushButton_showSummary"] = "Show a textual summary of your model."
    tooltips["label_inpImgSize"] = "Input image size."
    tooltips["label_outpSize"] = "Output dimension of model. Typically, the number of classes."
    tooltips["pushButton_toTensorB"] = "Show model architecture in Tensorboard (web browser will be started)."
    tooltips["groupBox_imageShow"] = "Our image will be shown here."
    tooltips["groupBox_image_Settings"] = "Plotting options."
    tooltips["label_image_alpha"] = "Alpha value for the image."
    tooltips["groupBox_gradCAM_Settings"] = "Settings for Grad-CAM activation heatmap. Title of original paper: Grad-CAM: Visual Explanations from Deep Networks via Gradient-based Localization; URL: https://arxiv.org/abs/1610.02391 "
    tooltips["label_gradCAM_targetClass"] = "Determine the class for which the activation heatmap should be shown."
    tooltips["label_gradCAM_targetLayer"] = "Determine the layer for which the activation heatmap should be show. Typically, the last convolutional layer is used."
    tooltips["label_gradCAM_colorMap"] = "Colormap for the Grad-CAM activation heatmap."
    tooltips["label_gradCAM_alpha"] = "Alpha value for the Grad-CAM heatmap."
    tooltips["pushButton_reset"] = "Reset settings."
    tooltips["pushButton_update"] = "Apply the settings and update the image."
    tooltips["groupBox_model"] = "Information about the model for which learning rates should be screened."
    tooltips["label_startLR"] = "Define the lower bound for screening learning rates."
    tooltips["label_stopLr"] = "Define the upper bound for screening learning rates."
    tooltips["label_percData"] = "Carrying out the screening for learning rates, requires to actually fit the model for some epochs. In case your loaded dataset is huge that might take very long. With this option here, you can choose to only use a subset of your data to speed up the process."
    tooltips["label_stepsPerEpoch"] = "The number of screening steps per epoch is calculated using nr_training_images / batch_size."
    tooltips["pushButton_LrReset"] = "Reset the LR screening settings to initial values."
    tooltips["pushButton_color"] = "Click to open a menu to choose a color for the LR screening plot."
    tooltips["label_lineWidth"] = "Define the width of the line in the LR screening plot."

    tooltips["checkBox_smooth"] = "Smooth values using an exponentially weighed running average."
    tooltips["checkBox_valMetrics"] = "Compute validation metrics after each batch update. This can be computationally VERY expensive, but allows you to plot the validation loss and validation accuracy. Only when checked, the above dropdown menu will show these options."


    tooltips["label_epochs_lrfind"] = "Number of epochs to train. Higher will give more precise result. Typically, 3-5 epochs are sufficient."
    tooltips["pushButton_LrFindRun"] = "Run a LR screening. This might take some minutes since data has to be loaded and a model is actually trained for some epochs. Note that all parameters (image augmentation etc) are used as indicated in the corresponding menus. Hence, make sure to first set these values as it affects training."
    tooltips["groupBox_LrSettings"] = "Carry out a screening of different learning rates as introduced in https://arxiv.org/abs/1506.01186.<br>Very low learning rates will not allow the model to learn anything and the resulting loss stays constant. Too high learning rates will cause large updates of model weights which prevents finding a minumum. You may consult this page to learn more how to choose good learning rates and how it is implemented:<br>https://www.pyimagesearch.com/2019/08/05/keras-learning-rate-finder "    
    tooltips["groupBox_singleLr"] = "When activating the box, a line will appear in the plot above, which you can drag to define a single learning rate value. You should choose a point where the loss is very low (minimum), but does yet start to fluctuate."
    tooltips["groupBox_LrRange"] = "When activating this box, a range will appear in the plot above, which you can drag to define a range of learning rates. The left edge should be located where the loss starts to decrease and the right edge close to the minumum."
    tooltips["pushButton_LR_plot"] = "Plot the learning rate vs epochs. The plot depends on the parameters defined in this menu. The plot allows you to decide whether you like the leanring rate schedule."

    tooltips["tableWidget_pop_pTr_layersTable"] = "The table shows all Dense and Conv2D layers and their trainablilities. To decrease the trainability, use the spinbox and a value between 0 and 1 and hit 'Update'. Then, the layers where you chose decrease the trainabliity are split into a layer that stays trainable and another layer that is set to not-trainable. Next, a Concatenation-layer is inserted, which combines both aforementioned layers again. Of course the weights are copied from the initial model, to the customized model, but you cannot specify, which particular nodes/filters of a layer are trainable or non-trainable. If you set for example a trainabliity of 0.25, the first 75% of the nodes are set to not-trainable and the rest 25% are left trainable."
    tooltips["pushButton_pop_pTr_update"] = "Apply the requested changes in trainability."
    tooltips["pushButton_pop_pTr_ok"] = "Save the customized model to a user-defined location. This model will automaticlly be selected for 'Load and continue' in the 'Define model'-tab. Just determine a 'Model path' before training. Training will then start using your customized model."

    #tooltips["msg_loadSession"] = "This function does only refurbish the window 'Drag and drop .rtdc files here'. To load the corresponding model please use 'Load and continue' in the 'Define Model' -tab. Image augmentation parameters have to be adjusted manually."
    tooltips["msg_loadSession"] = "Should only be the box 'Drag and drop .rtdc files here' be loaded, or the full set of hyper-parameters (model type, image augmentation parameters, learning rate,...)?"

    tooltips["label_saveMetaEvery"] = "Determine after every how many seconds the meta.xlsx file should be saved. For training procedures with more than 10k epochs saving of the meta.xlsx file might take long and you want to reduce the saving frequency. Also, saving the meta.xlsx file is only important for safety (keep progress also in case of power outage) and if you want to regularly check the excel file (e.g. using the 'History'-Tab)."









    ####deprecated!
    tooltips["checkBox_learningRate"] = "Change the learning rate. The default optimizer is \'adam\' with a learning rate of 0.001"


    for key in tooltips.keys():
        tooltips[key] = "<html><head/><body><p>"+tooltips[key]+"</p></body></html>"
    
    
    return tooltips



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 22 16:20:53 2022

@author: nana
"""
import os,shutil,json,re,urllib
import h5py, hashlib, warnings, pathlib
import datetime, time
import six, zipfile, tarfile
import numpy as np
from scipy.stats import gaussian_kde, skew
from scipy.interpolate import RectBivariateSpline

from . import aid_img

dir_root = os.path.dirname(__file__)#ask the module for its origin

#: Chunk size for storing HDF5 data
CHUNK_SIZE = 100

def load_rtdc(rtdc_path):
    """
    This function load .rtdc files using h5py and takes care of catching all
    errors
    """
    try:
        try:
            #sometimes there occurs an error when opening hdf files,
            #therefore try opening a second time in case of an error.
            #This is very strange, and seems like a dirty solution,
            #but I never saw it failing two times in a row
            rtdc_ds = h5py.File(rtdc_path, 'r')
        except:
            rtdc_ds = h5py.File(rtdc_path, 'r')
        return False,rtdc_ds #failed=False
    except Exception as e:
        #There is an issue loading the files!
        return True,e

def save_aid_settings(Default_dict):
    dir_settings = os.path.join(dir_root,"aid_settings.json")#dir to settings
    #Save the layout to Default_dict
    with open(dir_settings, 'w') as f:
        json.dump(Default_dict,f)

def hashfile(fname, blocksize=65536, count=0, constructor=hashlib.md5,
             hasher_class=None):
    """Compute md5 hex-hash of a file
    Parameters
    ----------
    fname: str or pathlib.Path
        path to the file
    blocksize: int
        block size in bytes read from the file
        (set to `0` to hash the entire file)
    count: int
        number of blocks read from the file
    hasher_class: callable
        deprecated, see use `constructor` instead
    constructor: callable
        hash algorithm constructor
    """
    if hasher_class is not None:
        warnings.warn("The `hasher_class` argument is deprecated, please use "
                      "`constructor` instead.")
        constructor = hasher_class
    hasher = constructor()
    fname = pathlib.Path(fname)
    with fname.open('rb') as fd:
        buf = fd.read(blocksize)
        ii = 0
        while len(buf) > 0:
            hasher.update(buf)
            buf = fd.read(blocksize)
            ii += 1
            if count and ii == count:
                break
    return hasher.hexdigest()

def obj2bytes(obj):
    """Bytes representation of an object for hashing"""
    if isinstance(obj, str):
        return obj.encode("utf-8")
    elif isinstance(obj, pathlib.Path):
        return obj2bytes(str(obj))
    elif isinstance(obj, (bool, int, float)):
        return str(obj).encode("utf-8")
    elif obj is None:
        return b"none"
    elif isinstance(obj, np.ndarray):
        return obj.tobytes()
    elif isinstance(obj, tuple):
        return obj2bytes(list(obj))
    elif isinstance(obj, list):
        return b"".join(obj2bytes(o) for o in obj)
    elif isinstance(obj, dict):
        return obj2bytes(sorted(obj.items()))
    elif hasattr(obj, "identifier"):
        return obj2bytes(obj.identifier)
    elif isinstance(obj, h5py.Dataset):
        return obj2bytes(obj[0])
    else:
        raise ValueError("No rule to convert object '{}' to string.".
                         format(obj.__class__))

def hashfunction(rtdc_path):
    """Hash value based on file name and content"""
    tohash = [os.path.basename(rtdc_path),
              # Hash a maximum of ~1MB of the hdf5 file
              hashfile(rtdc_path, blocksize=65536, count=20)]
    return hashlib.md5(obj2bytes(tohash)).hexdigest()

def calc_ram_need(crop):
    crop = int(crop)
    n=1000
    #100k cropped images (64x64 pix unsigned integer8) need 400MB of RAM:
    Imgs = []
    sizex,sizey = crop,crop
    for i in range(n):
        Imgs.append(np.random.randint(low=0,high=255,size=(sizex,sizey),dtype=np.uint8))
    Imgs = np.array(Imgs)
    #Imgs = Imgs.astype(np.uint8)
    MB = Imgs.nbytes/1048576.0 #Amount of RAM for 1000 images
    MB = MB/float(n)
    return MB

def ram_compare_data(ram_dic,new_dic):    
    #compare the rtdc filenames:
    new_rtdc_paths = [a["rtdc_path"] for a in new_dic["SelectedFiles"]]
    ram_rtdc_paths = list(ram_dic["rtdc_path"])
    test_rtdc_paths = set(ram_rtdc_paths)==set(new_rtdc_paths)

    #Compare the image shape (size)
    ram_imgshape = ram_dic["Cropped_Images"][0].shape
    ram_imgcrop = ram_imgshape[1]
    new_imgcrop = new_dic["cropsize2"]
    test_imgcrop = ram_imgcrop==new_imgcrop
    
    #Compare the colormode
    if len(ram_imgshape)==3:
        ram_colormode = "grayscale"
    elif len(ram_imgshape)==4 and ram_imgshape[-1]==3:
        ram_colormode = "rgb"
    else:
        print("Image dimension not supported")
    new_colormode = new_dic["color_mode"].lower()
    test_colormode = ram_colormode==new_colormode
      
    #compare the number of images
    ram_nr_images = [a.shape[0] for a in ram_dic["Cropped_Images"]]
    new_nr_images = [a["nr_images"] for a in new_dic["SelectedFiles"]]
    test_nr_images = set(ram_nr_images)==set(new_nr_images)
 
    dic = {"test_rtdc_paths":test_rtdc_paths,"test_imgcrop":test_imgcrop,"test_colormode":test_colormode,"test_nr_images":test_nr_images}
    #Are all tests poisitve (True)?
    alltrue = all(dic.values())
    return alltrue


def write_rtdc(fname,rtdc_datasets,X_valid,Indices,cropped=True,color_mode='Grayscale',xtra_in=[]):
    """
    fname - path+filename of file to be created
    rtdc_datasets - list paths to rtdc data-data-sets
    X_valid - list containing numpy arrays. Each array contains cropped images of individual measurements corresponding to each rtdc_ds
    Indices - list containing numpy arrays. Each array contais index values which refer to the index of the cell in the original rtdc_ds
    """
    #Check if a file with name fname already exists:
    if os.path.isfile(fname):
        os.remove(fname) #delete it
        print("overwrite existing file")
    
    index_new = np.array(range(1,int(np.sum(np.array([len(I) for I in Indices]))+1))) #New index. Will replace the existing index in order to support viewing imges in shapeout   

    #quickly check the image-format of the validation data:
    images_shape = []
    for i in range(len(rtdc_datasets)):
        failed,rtdc_ds = load_rtdc(rtdc_datasets[i])
        if failed:
            print("Error occurred during loading file\n"+str(rtdc_datasets[i])+"\n"+str(rtdc_ds))
        else:
            #get the shape for all used files
            images_shape.append(rtdc_ds["events"]["image"].shape) 
    #Allow RGB export only, of all files have 3 channels
    if cropped==False: #only if the original images should be exported
        #the length of the images is shape=3 for grayscale images and =4 for RGB images
        images_shape_l = [len(images_shape[i]) for i in range(len(images_shape))]
        channels = [s[-1] for s in images_shape] #get the nr of channels in each file
        if images_shape_l==len(images_shape_l)*[4]: #if all files have len(shape)=4
            if channels==len(channels)*[3]: #check if all files have 3 channels 
                print("All files have three channels. I will save the entire RGB information")
                color_mode = "RGB"
            else: #(some) files neither have 1 nor 3 channels. Not supported
                print("AID currently only supports single channel (grayscale) and three channel (RGB) data. Your data has following channels: "+str(channels))
                return
                #TODO: Maybe support more channels
        else: #not all files have three channels->ergo there is no way to keep the color-information:
            color_mode = "Grayscale"

    #########Only for RGB Images!: Collect all data first and write############ 
    if color_mode=='RGB': #or images_shape_max==4:
    #if len(X_valid)>0 and len(np.array(X_valid).shape)==5: #RGB image
        #Get all images,pos_x,pos_y:
        images,pos_x,pos_y = [],[],[]
        for i in range(len(rtdc_datasets)):
            failed,rtdc_ds = load_rtdc(rtdc_datasets[i])
            if failed:
                print("Error occurred during loading file\n"+str(rtdc_datasets[i])+"\n"+str(rtdc_ds))
            else:
                indices = Indices[i]
                if len(indices>0):
                    if cropped==False:
                        images.append([rtdc_ds["events"]["image"][ii] for ii in indices])
                    pos_x.append([rtdc_ds["events"]["pos_x"][ii] for ii in indices])
                    pos_y.append([rtdc_ds["events"]["pos_y"][ii] for ii in indices])
                    
        if cropped==True:
            images = X_valid
            #If the user want to export cropped images, then they will be
            #in the color mode of the model

        images = np.concatenate(images)
        pos_x = np.concatenate(pos_x)
        pos_y = np.concatenate(pos_y)
        
        #copy the empty Empty.rtdc
        shutil.copy(os.path.join(dir_root,"Empty.rtdc"),fname)
        
        maxshape = (None, images.shape[1], images.shape[2], images.shape[3])
        #Create rtdc_dataset
        hdf = h5py.File(fname,'a')
        hdf.create_dataset("events/image", data=images, dtype=np.uint8,maxshape=maxshape,fletcher32=True,chunks=True)
        hdf.create_dataset("events/pos_x", data=pos_x, dtype=np.int32)
        hdf.create_dataset("events/pos_y", data=pos_y, dtype=np.int32)
        hdf.create_dataset("events/index", data=index_new, dtype=np.int32)
        if len((np.array(xtra_in)).ravel())>0: #
            #hdf.create_dataset("xtra_in", data=xtra_in, dtype=np.float32)
            hdf.create_dataset('xtra_in', data=np.concatenate(xtra_in), compression="gzip", chunks=True, maxshape=(None,))
        
        #Adjust metadata:
        #"experiment:event count" = Nr. of images
        hdf.attrs["experiment:event count"]=images.shape[0]
        hdf.attrs["experiment:sample"]=fname
        hdf.attrs["imaging:pixel size"]=rtdc_ds.attrs["imaging:pixel size"]
        hdf.close()
        return

    Features,Trace_lengths,Mask_dims_x,Mask_dims_y,Img_dims_x,Img_dims_y = [],[],[],[],[],[]
    for i in range(len(rtdc_datasets)):
        failed,rtdc_ds = load_rtdc(rtdc_datasets[i])
        if failed:
            print("Error occurred during loading file\n"+str(rtdc_datasets[i])+"\n"+str(rtdc_ds))
        else:
            features = list(rtdc_ds["events"].keys())#rtdc_ds._events.keys()#all features
            Features.append(features)
    
            #The lengths of the fluorescence traces have to be equal, otherwise those traces also have to be dropped
            if "trace" in features:
                trace_lengths = [(rtdc_ds["events"]["trace"][tr][0]).size for tr in rtdc_ds["events"]["trace"].keys()]
                Trace_lengths.append(trace_lengths)
            #Mask Image dimensions have to be equal, otherwise those mask have to be dropped
            if "mask" in features:
                mask_dim = (rtdc_ds["events"]["mask"][0]).shape
                Mask_dims_x.append(mask_dim[0])
                Mask_dims_y.append(mask_dim[1])
            #Mask Image dimensions have to be equal, otherwise those mask have to be dropped
            if "image" in features:
                img_dim = (rtdc_ds["events"]["image"][0]).shape
                Img_dims_x.append(img_dim[0])
                Img_dims_y.append(img_dim[1])
 
    #Find common features in all .rtdc sets:
    def commonElements(arr): 
        # initialize result with first array as a set 
        result = set(arr[0]) 
        for currSet in arr[1:]: 
            result.intersection_update(currSet) 
        return list(result)     
    features = commonElements(Features)

    # features = ["index_orig"] + features
    # if "index" not in features:
    #     features = ["index"] + features
        
    if "trace" in features:
        Trace_lengths = np.concatenate(Trace_lengths)
        trace_lengths = np.unique(np.array(Trace_lengths))            
        if len(trace_lengths)>1:
            ind = np.where(np.array(features)!="trace")[0]
            features = list(np.array(features)[ind])
            print("Dropped traces becasue of unequal lengths")

    if "mask" in features:
        mask_dim_x = np.unique(np.array(Mask_dims_x))            
        mask_dim_y = np.unique(np.array(Mask_dims_y))            
        if len(mask_dim_x)>1 or len(mask_dim_y)>1:
            ind = np.where(np.array(features)!="mask")[0]
            features = list(np.array(features)[ind])
            print("Dropped mask becasue of unequal image sizes")

    if "image" in features:
        img_dim_x = np.unique(np.array(Img_dims_x))            
        img_dim_y = np.unique(np.array(Img_dims_y))            
        if len(img_dim_x)>1 or len(img_dim_y)>1:
            print("Unequal image dimensions -> Force export of cropped images!")
            cropped = True

    for i in range(len(rtdc_datasets)):
        failed,rtdc_ds = load_rtdc(rtdc_datasets[i])
        if failed:
            print("Error occurred during loading file\n"+str(rtdc_datasets[i])+"\n"+str(rtdc_ds))
        else:
            indices = Indices[i]
            Images = X_valid[i]
            
            if len(Images)>0 and len(Images.shape)==3:
                index_new_ = index_new[0:len(indices)]
                index_new = np.delete(index_new,range(len(indices)))

                h5obj = h5py.File(fname,'a')
                events = h5obj.require_group("events")
                
                # write each feature individually
                for feat in features:
                    if feat == "contour":
                        print("Omitting contours")
                        #cont_list = [rtdc_ds["events"]["contour"][ii] for ii in indices]
                        #store_contour(h5group=events,name=feat,data=cont_list, compression="gzip")

                    # elif feat == "index_orig":
                    #     store_scalar(h5group=events,name=feat, data=index_new_, compression="gzip")

                    # elif feat == "index_orig":
                    #     values = np.array(rtdc_ds["events"]["index"])[indices]
                    #     store_scalar(h5group=events,name=feat, data=values, compression="gzip")
                    
                    elif feat=="mask":# in ["mask", "image"]:
                        mask = np.array(rtdc_ds["events"]["mask"])[indices]
                        if cropped:
                            pix = rtdc_ds.attrs["imaging:pixel size"]
                            pos_x = np.array(rtdc_ds["events"]["pos_x"])[indices]/pix
                            pos_y = np.array(rtdc_ds["events"]["pos_y"])[indices]/pix
                            img_dim_x = Images[0].shape[1]
                            img_dim_y = Images[0].shape[0]
                            mask = aid_img.image_crop_pad_cv2(list(mask),pos_x,pos_y,pix,img_dim_x,img_dim_y,padding_mode="cv2.BORDER_CONSTANT")
                        
                        mask = np.asarray(mask, dtype=np.uint8)
                        if mask.max() != 255 and mask.max() != 0 and mask.min() == 0:
                            mask = mask / mask.max() * 255
                        maxshape = (None, mask.shape[1], mask.shape[2])
                        store_image(h5group=events,name=feat, data=mask, compression="gzip")
                        # dset = h5obj.create_dataset("events/"+feat, data=mask, dtype=np.uint8,maxshape=maxshape,fletcher32=True,chunks=True)
                        # dset.attrs.create('CLASS', np.string_('IMAGE'))
                        # dset.attrs.create('IMAGE_VERSION', np.string_('1.2'))
                        # dset.attrs.create('IMAGE_SUBCLASS', np.string_('IMAGE_GRAYSCALE'))
                        #store_mask(h5group=events,name=feat, data=mask, compression="gzip")

                    elif "image" in feat:
                        if cropped and feat=="image":
                            image_list = np.array(Images)
                        elif cropped and feat != "image":#there are images of another channel, they need ot be cropped first
                            pix = rtdc_ds.attrs["imaging:pixel size"]
                            pos_x = np.array(rtdc_ds["events"]["pos_x"])[indices]/pix
                            pos_y = np.array(rtdc_ds["events"]["pos_y"])[indices]/pix
                            img_dim_x = Images[0].shape[1]
                            img_dim_y = Images[0].shape[0]
                            image_list = np.array(rtdc_ds["events"][feat])[indices]
                            image_list = aid_img.image_crop_pad_cv2(list(image_list),pos_x,pos_y,pix,img_dim_x,img_dim_y,padding_mode="cv2.BORDER_CONSTANT")
                            image_list = np.array(image_list)
                        else:
                            image_list = np.array(rtdc_ds["events"][feat])[indices]
                        store_image(h5group=events,name=feat, data=image_list, compression="gzip")

                    elif feat == "trace":
                        # create events group
                        data = rtdc_ds["events"]["trace"]
                        h5_group = h5obj["events"].require_group("trace")
                        store_trace(h5_group=h5_group,data=data,indices=indices)
                        #dclab.rtdc_dataset.write_hdf5.store_trace(h5group=events,name=feat,data=trace,compression="gzip")

                    elif feat == "pos_x" and cropped==True:
                        values = np.zeros(shape=len(indices))+np.round(img_dim_x/2.0)*rtdc_ds.attrs["imaging:pixel size"]
                        store_scalar(h5group=events, name=feat, data=values, compression="gzip")
                        #dclab.rtdc_dataset.write_hdf5.store_scalar(h5group=events, name=feat, data=values, compression="gzip")

                    elif feat == "pos_y" and cropped==True:
                        values = np.zeros(shape=len(indices))+np.round(img_dim_y/2.0)*rtdc_ds.attrs["imaging:pixel size"]
                        store_scalar(h5group=events, name=feat, data=values, compression="gzip")
                        
                    else:
                        values = np.array(rtdc_ds["events"][feat])[indices]
                        store_scalar(h5group=events, name=feat, data=values, compression="gzip")

                #Adjust metadata:
                #"experiment:event count" = Nr. of images
                h5obj.attrs["experiment:event count"] = np.sum(np.array([len(indi) for indi in Indices])) 
                if cropped:
                    #Adjust the meta for cropped images
                    img_dim_x = Images[0].shape[1]
                    img_dim_y = Images[0].shape[0]
                    h5obj.attrs["imaging:roi size x"] = img_dim_x
                    h5obj.attrs["imaging:roi size y"] = img_dim_y

                meta_keys = list(rtdc_ds.attrs)
                if "experiment:date" not in meta_keys:
                    h5obj.attrs["experiment:date"] = time.strftime("%Y-%m-%d")
                if "experiment:time" not in meta_keys:
                    h5obj.attrs["experiment:time"] = time.strftime("%Y-%m-%d")
                    
                for meta_key in meta_keys:
                    h5obj.attrs[meta_key] = rtdc_ds.attrs[meta_key]

                h5obj.close()
                    
    #Append xtra_in data to the rtdc file
    if len((np.array(xtra_in)).ravel())>0: #in case there is some xtra_in data
        with h5py.File(fname, 'a') as rtdc_h5:
            try:
                rtdc_h5.create_dataset('xtra_in', data=np.concatenate(xtra_in), compression="gzip", chunks=True, maxshape=(None,))
                rtdc_h5.close()
            except:
                pass
            
def store_image(h5group, name, data, compression, background=False):
    """Store image data in an HDF5 group
    Parameters
    ----------
    h5group: h5py.Group
        The group (usually "events") where to store the image data
    data: 2d or 3d ndarray
        The image data. If 3d, then the first axis enumerates
        the images.
    compression: str
        Dataset compression method
    background: bool
        If set to False (default), then the regular "image" is stored;
        If set to True, then the background image ("image_bg") is
        stored.
    """
    if len(data.shape) == 2:
        # single event
        data = data.reshape(1, data.shape[0], data.shape[1])
    if name not in h5group:
        maxshape = (None, data.shape[1], data.shape[2])
        chunks = (CHUNK_SIZE, data.shape[1], data.shape[2])
        dset = h5group.create_dataset(name,
                                      data=data,
                                      dtype=np.uint8,
                                      maxshape=maxshape,
                                      chunks=chunks,
                                      fletcher32=True,
                                      compression=compression)
        # Create and Set image attributes:
        # HDFView recognizes this as a series of images.
        # Use np.string_ as per
        # http://docs.h5py.org/en/stable/strings.html#compatibility
        dset.attrs.create('CLASS', np.string_('IMAGE'))
        dset.attrs.create('IMAGE_VERSION', np.string_('1.2'))
        dset.attrs.create('IMAGE_SUBCLASS', np.string_('IMAGE_GRAYSCALE'))
    else:
        dset = h5group[name]
        oldsize = dset.shape[0]
        dset.resize(oldsize + data.shape[0], axis=0)
        dset[oldsize:] = data


def store_mask(h5group,name, data, compression):
    # store binary mask data as uint8 to allow visualization in HDFView
    data = np.asarray(data, dtype=np.uint8)
    if data.max() != 255 and data.max() != 0 and data.min() == 0:
        data = data / data.max() * 255
    if len(data.shape) == 2:
        # single event
        data = data.reshape(1, data.shape[0], data.shape[1])
    if "mask" not in h5group:
        maxshape = (None, data.shape[1], data.shape[2])
        chunks = (CHUNK_SIZE, data.shape[1], data.shape[2])
        dset = h5group.create_dataset(name,
                                      data=data,
                                      dtype=np.uint8,
                                      maxshape=maxshape,
                                      chunks=chunks,
                                      fletcher32=True,
                                      compression=compression)
        # Create and Set image attributes
        # HDFView recognizes this as a series of images
        dset.attrs.create('CLASS', np.string_('IMAGE'))
        dset.attrs.create('IMAGE_VERSION', np.string_('1.2'))
        dset.attrs.create('IMAGE_SUBCLASS', np.string_('IMAGE_GRAYSCALE'))
    else:
        dset = h5group[name]
        oldsize = dset.shape[0]
        dset.resize(oldsize + data.shape[0], axis=0)
        dset[oldsize:] = data


def store_scalar(h5group, name, data, compression):
    if np.isscalar(data):
        # single event
        data = np.atleast_1d(data)
    if name not in h5group:
        h5group.create_dataset(name,
                               data=data,
                               maxshape=(None,),
                               chunks=(CHUNK_SIZE,),
                               fletcher32=True,
                               compression=compression
                               )
    else:
        dset = h5group[name]
        oldsize = dset.shape[0]
        dset.resize(oldsize + data.shape[0], axis=0)
        dset[oldsize:] = data


def store_trace(h5_group, data, indices):
    keys_trace = list(data.keys())
    # create trace group
    #grp = h5obj.require_group("trace")

    for flt in keys_trace:
        # create traces datasets
        if flt not in h5_group:
            values = np.array(data[flt][:])
            values = values[indices]
            if len(values) == 1:# single event
                values = values.reshape(1, -1)
            maxshape = (None, values.shape[1])
            chunks = (CHUNK_SIZE, values.shape[1])
            h5_group.create_dataset(flt,data=values,maxshape=maxshape,
                               chunks=chunks,fletcher32=True,compression="gzip")
        else:
            dset = h5_group[flt]
            oldsize = dset.shape[0]
            values = np.array(data[flt][:])
            values = values[indices]
            dset.resize(oldsize + values.shape[0], axis=0)
            dset[oldsize:] = values


def write_rtdc(fname,rtdc_datasets,X_valid,Indices,cropped=True,color_mode='Grayscale',xtra_in=[]):
    """
    fname - path+filename of file to be created
    rtdc_datasets - list paths to rtdc data-data-sets
    X_valid - list containing numpy arrays. Each array contains cropped images of individual measurements corresponding to each rtdc_ds
    Indices - list containing numpy arrays. Each array contais index values which refer to the index of the cell in the original rtdc_ds
    """
    #Check if a file with name fname already exists:
    if os.path.isfile(fname):
        os.remove(fname) #delete it
        print("overwrite existing file")
    
    index_new = np.array(range(1,int(np.sum(np.array([len(I) for I in Indices]))+1))) #New index. Will replace the existing index in order to support viewing imges in shapeout   

    #quickly check the image-format of the validation data:
    images_shape = []
    for i in range(len(rtdc_datasets)):
        failed,rtdc_ds = load_rtdc(rtdc_datasets[i])
        if failed:
            print("Error occurred during loading file\n"+str(rtdc_datasets[i])+"\n"+str(rtdc_ds))
        else:
            #get the shape for all used files
            images_shape.append(rtdc_ds["events"]["image"].shape) 
    #Allow RGB export only, of all files have 3 channels
    if cropped==False: #only if the original images should be exported
        #the length of the images is shape=3 for grayscale images and =4 for RGB images
        images_shape_l = [len(images_shape[i]) for i in range(len(images_shape))]
        channels = [s[-1] for s in images_shape] #get the nr of channels in each file
        if images_shape_l==len(images_shape_l)*[4]: #if all files have len(shape)=4
            if channels==len(channels)*[3]: #check if all files have 3 channels 
                print("All files have three channels. I will save the entire RGB information")
                color_mode = "RGB"
            else: #(some) files neither have 1 nor 3 channels. Not supported
                print("AID currently only supports single channel (grayscale) and three channel (RGB) data. Your data has following channels: "+str(channels))
                return
                #TODO: Maybe support more channels
        else: #not all files have three channels->ergo there is no way to keep the color-information:
            color_mode = "Grayscale"

    #########Only for RGB Images!: Collect all data first and write############ 
    if color_mode=='RGB': #or images_shape_max==4:
    #if len(X_valid)>0 and len(np.array(X_valid).shape)==5: #RGB image
        #Get all images,pos_x,pos_y:
        images,pos_x,pos_y = [],[],[]
        for i in range(len(rtdc_datasets)):
            failed,rtdc_ds = load_rtdc(rtdc_datasets[i])
            if failed:
                print("Error occurred during loading file\n"+str(rtdc_datasets[i])+"\n"+str(rtdc_ds))
            else:
                indices = Indices[i]
                if len(indices>0):
                    if cropped==False:
                        images.append([rtdc_ds["events"]["image"][ii] for ii in indices])
                    pos_x.append([rtdc_ds["events"]["pos_x"][ii] for ii in indices])
                    pos_y.append([rtdc_ds["events"]["pos_y"][ii] for ii in indices])
                    
        if cropped==True:
            images = X_valid
            #If the user want to export cropped images, then they will be
            #in the color mode of the model

        images = np.concatenate(images)
        pos_x = np.concatenate(pos_x)
        pos_y = np.concatenate(pos_y)
        
        #copy the empty Empty.rtdc
        shutil.copy(os.path.join(dir_root,"Empty.rtdc"),fname)
        
        maxshape = (None, images.shape[1], images.shape[2], images.shape[3])
        #Create rtdc_dataset
        hdf = h5py.File(fname,'a')
        hdf.create_dataset("events/image", data=images, dtype=np.uint8,maxshape=maxshape,fletcher32=True,chunks=True)
        hdf.create_dataset("events/pos_x", data=pos_x, dtype=np.int32)
        hdf.create_dataset("events/pos_y", data=pos_y, dtype=np.int32)
        hdf.create_dataset("events/index", data=index_new, dtype=np.int32)
        if len((np.array(xtra_in)).ravel())>0: #
            #hdf.create_dataset("xtra_in", data=xtra_in, dtype=np.float32)
            hdf.create_dataset('xtra_in', data=np.concatenate(xtra_in), compression="gzip", chunks=True, maxshape=(None,))
        
        #Adjust metadata:
        #"experiment:event count" = Nr. of images
        hdf.attrs["experiment:event count"]=images.shape[0]
        hdf.attrs["experiment:sample"]=fname
        hdf.attrs["imaging:pixel size"]=rtdc_ds.attrs["imaging:pixel size"]
        hdf.close()
        return

    Features,Trace_lengths,Mask_dims_x,Mask_dims_y,Img_dims_x,Img_dims_y = [],[],[],[],[],[]
    for i in range(len(rtdc_datasets)):
        failed,rtdc_ds = load_rtdc(rtdc_datasets[i])
        if failed:
            print("Error occurred during loading file\n"+str(rtdc_datasets[i])+"\n"+str(rtdc_ds))
        else:
            features = list(rtdc_ds["events"].keys())#rtdc_ds._events.keys()#all features
            Features.append(features)
    
            #The lengths of the fluorescence traces have to be equal, otherwise those traces also have to be dropped
            if "trace" in features:
                trace_lengths = [(rtdc_ds["events"]["trace"][tr][0]).size for tr in rtdc_ds["events"]["trace"].keys()]
                Trace_lengths.append(trace_lengths)
            #Mask Image dimensions have to be equal, otherwise those mask have to be dropped
            if "mask" in features:
                mask_dim = (rtdc_ds["events"]["mask"][0]).shape
                Mask_dims_x.append(mask_dim[0])
                Mask_dims_y.append(mask_dim[1])
            #Mask Image dimensions have to be equal, otherwise those mask have to be dropped
            if "image" in features:
                img_dim = (rtdc_ds["events"]["image"][0]).shape
                Img_dims_x.append(img_dim[0])
                Img_dims_y.append(img_dim[1])
 
    #Find common features in all .rtdc sets:
    def commonElements(arr): 
        # initialize result with first array as a set 
        result = set(arr[0]) 
        for currSet in arr[1:]: 
            result.intersection_update(currSet) 
        return list(result)     
    features = commonElements(Features)

    # features = ["index_orig"] + features
    # if "index" not in features:
    #     features = ["index"] + features
        
    if "trace" in features:
        Trace_lengths = np.concatenate(Trace_lengths)
        trace_lengths = np.unique(np.array(Trace_lengths))            
        if len(trace_lengths)>1:
            ind = np.where(np.array(features)!="trace")[0]
            features = list(np.array(features)[ind])
            print("Dropped traces becasue of unequal lengths")

    if "mask" in features:
        mask_dim_x = np.unique(np.array(Mask_dims_x))            
        mask_dim_y = np.unique(np.array(Mask_dims_y))            
        if len(mask_dim_x)>1 or len(mask_dim_y)>1:
            ind = np.where(np.array(features)!="mask")[0]
            features = list(np.array(features)[ind])
            print("Dropped mask becasue of unequal image sizes")

    if "image" in features:
        img_dim_x = np.unique(np.array(Img_dims_x))            
        img_dim_y = np.unique(np.array(Img_dims_y))            
        if len(img_dim_x)>1 or len(img_dim_y)>1:
            print("Unequal image dimensions -> Force export of cropped images!")
            cropped = True

    for i in range(len(rtdc_datasets)):
        failed,rtdc_ds = load_rtdc(rtdc_datasets[i])
        if failed:
            print("Error occurred during loading file\n"+str(rtdc_datasets[i])+"\n"+str(rtdc_ds))
        else:
            indices = Indices[i]
            Images = X_valid[i]
            
            if len(Images)>0 and len(Images.shape)==3:
                index_new_ = index_new[0:len(indices)]
                index_new = np.delete(index_new,range(len(indices)))

                h5obj = h5py.File(fname,'a')
                events = h5obj.require_group("events")
                
                # write each feature individually
                for feat in features:
                    if feat == "contour":
                        print("Omitting contours")
                        #cont_list = [rtdc_ds["events"]["contour"][ii] for ii in indices]
                        #store_contour(h5group=events,name=feat,data=cont_list, compression="gzip")

                    # elif feat == "index_orig":
                    #     store_scalar(h5group=events,name=feat, data=index_new_, compression="gzip")

                    # elif feat == "index_orig":
                    #     values = np.array(rtdc_ds["events"]["index"])[indices]
                    #     store_scalar(h5group=events,name=feat, data=values, compression="gzip")
                    
                    elif feat=="mask":# in ["mask", "image"]:
                        mask = np.array(rtdc_ds["events"]["mask"])[indices]
                        if cropped:
                            pix = rtdc_ds.attrs["imaging:pixel size"]
                            pos_x = np.array(rtdc_ds["events"]["pos_x"])[indices]/pix
                            pos_y = np.array(rtdc_ds["events"]["pos_y"])[indices]/pix
                            img_dim_x = Images[0].shape[1]
                            img_dim_y = Images[0].shape[0]
                            mask = aid_img.image_crop_pad_cv2(list(mask),pos_x,pos_y,pix,img_dim_x,img_dim_y,padding_mode="cv2.BORDER_CONSTANT")
                        
                        mask = np.asarray(mask, dtype=np.uint8)
                        if mask.max() != 255 and mask.max() != 0 and mask.min() == 0:
                            mask = mask / mask.max() * 255
                        maxshape = (None, mask.shape[1], mask.shape[2])
                        store_image(h5group=events,name=feat, data=mask, compression="gzip")
                        # dset = h5obj.create_dataset("events/"+feat, data=mask, dtype=np.uint8,maxshape=maxshape,fletcher32=True,chunks=True)
                        # dset.attrs.create('CLASS', np.string_('IMAGE'))
                        # dset.attrs.create('IMAGE_VERSION', np.string_('1.2'))
                        # dset.attrs.create('IMAGE_SUBCLASS', np.string_('IMAGE_GRAYSCALE'))
                        #store_mask(h5group=events,name=feat, data=mask, compression="gzip")

                    elif "image" in feat:
                        if cropped and feat=="image":
                            image_list = np.array(Images)
                        elif cropped and feat != "image":#there are images of another channel, they need ot be cropped first
                            pix = rtdc_ds.attrs["imaging:pixel size"]
                            pos_x = np.array(rtdc_ds["events"]["pos_x"])[indices]/pix
                            pos_y = np.array(rtdc_ds["events"]["pos_y"])[indices]/pix
                            img_dim_x = Images[0].shape[1]
                            img_dim_y = Images[0].shape[0]
                            image_list = np.array(rtdc_ds["events"][feat])[indices]
                            image_list = aid_img.image_crop_pad_cv2(list(image_list),pos_x,pos_y,pix,img_dim_x,img_dim_y,padding_mode="cv2.BORDER_CONSTANT")
                            image_list = np.array(image_list)
                        else:
                            image_list = np.array(rtdc_ds["events"][feat])[indices]
                        store_image(h5group=events,name=feat, data=image_list, compression="gzip")

                    elif feat == "trace":
                        # create events group
                        data = rtdc_ds["events"]["trace"]
                        h5_group = h5obj["events"].require_group("trace")
                        store_trace(h5_group=h5_group,data=data,indices=indices)
                        #dclab.rtdc_dataset.write_hdf5.store_trace(h5group=events,name=feat,data=trace,compression="gzip")

                    elif feat == "pos_x" and cropped==True:
                        values = np.zeros(shape=len(indices))+np.round(img_dim_x/2.0)*rtdc_ds.attrs["imaging:pixel size"]
                        store_scalar(h5group=events, name=feat, data=values, compression="gzip")
                        #dclab.rtdc_dataset.write_hdf5.store_scalar(h5group=events, name=feat, data=values, compression="gzip")

                    elif feat == "pos_y" and cropped==True:
                        values = np.zeros(shape=len(indices))+np.round(img_dim_y/2.0)*rtdc_ds.attrs["imaging:pixel size"]
                        store_scalar(h5group=events, name=feat, data=values, compression="gzip")
                        
                    else:
                        values = np.array(rtdc_ds["events"][feat])[indices]
                        store_scalar(h5group=events, name=feat, data=values, compression="gzip")

                #Adjust metadata:
                #"experiment:event count" = Nr. of images
                h5obj.attrs["experiment:event count"] = np.sum(np.array([len(indi) for indi in Indices])) 
                if cropped:
                    #Adjust the meta for cropped images
                    img_dim_x = Images[0].shape[1]
                    img_dim_y = Images[0].shape[0]
                    h5obj.attrs["imaging:roi size x"] = img_dim_x
                    h5obj.attrs["imaging:roi size y"] = img_dim_y

                meta_keys = list(rtdc_ds.attrs)
                if "experiment:date" not in meta_keys:
                    h5obj.attrs["experiment:date"] = time.strftime("%Y-%m-%d")
                if "experiment:time" not in meta_keys:
                    h5obj.attrs["experiment:time"] = time.strftime("%Y-%m-%d")
                    
                for meta_key in meta_keys:
                    h5obj.attrs[meta_key] = rtdc_ds.attrs[meta_key]

                h5obj.close()
                    
    #Append xtra_in data to the rtdc file
    if len((np.array(xtra_in)).ravel())>0: #in case there is some xtra_in data
        with h5py.File(fname, 'a') as rtdc_h5:
            try:
                rtdc_h5.create_dataset('xtra_in', data=np.concatenate(xtra_in), compression="gzip", chunks=True, maxshape=(None,))
                rtdc_h5.close()
            except:
                pass


def create_temp_folder():
    temp_path = os.path.join(dir_root,"temp")
    if os.path.exists(temp_path):
        print("Found existing temporary folder: "+temp_path)
#        print("Delete all contents of that folder")
#        shutil.rmtree(temp_path,ignore_errors=True)
#        time.sleep(0.5)
#        try:
#            os.mkdir(temp_path)
#        except:
#            print("Could not delete temp folder. Files probably still in use!")
#            pass

    if not os.path.exists(temp_path):
        os.mkdir(temp_path)
        print("Created temporary folder: "+temp_path)
    return temp_path


#def delete_temp_folder():
#    temp_path = os.path.join(dir_root,"temp")
#    if os.path.exists(temp_path):
#        print("Found existing temporary folder: "+temp_path)
#        print("Delete all contents of that folder")
#        shutil.rmtree(temp_path,ignore_errors=True)
#        time.sleep(0.5)
#        try:
#            os.mkdir(temp_path)
#        except:
#            print("Could not delete temp folder. Files probably still in use!")
#            pass
#
#    if not os.path.exists(temp_path):#temp folder does not exist
#        os.mkdir(temp_path)#create the folder
#        print("Created temporary folder: "+temp_path)
#    return temp_path

def count_temp_folder():
    """
    Count the number of folders within the temporary folder
    """
    temp_path = os.path.join(dir_root,"temp")
    if os.path.exists(temp_path):#if temp even exists...
        files = os.listdir(temp_path)
        nr_temp_files = len(files)
    else:
        #there is no temp folder!
        create_temp_folder()#create a temp folder
        nr_temp_files = 0
    return nr_temp_files 

def open_temp():
    temp_path = create_temp_folder()
    os.startfile(temp_path)


def ram_compare_data(ram_dic,new_dic):    
    #compare the rtdc filenames:
    new_rtdc_paths = [a["rtdc_path"] for a in new_dic["SelectedFiles"]]
    ram_rtdc_paths = list(ram_dic["rtdc_path"])
    test_rtdc_paths = set(ram_rtdc_paths)==set(new_rtdc_paths)

    #Compare the image shape (size)
    ram_imgshape = ram_dic["Cropped_Images"][0].shape
    ram_imgcrop = ram_imgshape[1]
    new_imgcrop = new_dic["cropsize2"]
    test_imgcrop = ram_imgcrop==new_imgcrop
    
    #Compare the colormode
    if len(ram_imgshape)==3:
        ram_colormode = "grayscale"
    elif len(ram_imgshape)==4 and ram_imgshape[-1]==3:
        ram_colormode = "rgb"
    else:
        print("Image dimension not supported")
    new_colormode = new_dic["color_mode"].lower()
    test_colormode = ram_colormode==new_colormode
      
    #compare the number of images
    ram_nr_images = [a.shape[0] for a in ram_dic["Cropped_Images"]]
    new_nr_images = [a["nr_images"] for a in new_dic["SelectedFiles"]]
    test_nr_images = set(ram_nr_images)==set(new_nr_images)
 
    dic = {"test_rtdc_paths":test_rtdc_paths,"test_imgcrop":test_imgcrop,"test_colormode":test_colormode,"test_nr_images":test_nr_images}
    #Are all tests poisitve (True)?
    alltrue = all(dic.values())
    return alltrue
    
def download_zip(url_zip,fpath):
    """
    Download a zip file
    Parameters
    ----------
    url: str, URL to the downloadable zip file (e.g. https://github.com/maikherbig/AIDeveloper/releases/download/1.0.1-update/AIDeveloper_1.0.1-update.zip)
    fpath: str, path to store the downloaded data locally
    """
    error_msg = 'URL fetch failure on {}: {} -- {}'
        
    try:
      try:
        six.moves.urllib.request.urlretrieve(url=url_zip,filename=fpath)
      except six.moves.urllib.error.HTTPError as e:
        raise Exception(error_msg.format(url_zip, e.code, e.msg))
      except six.moves.urllib.error.URLError as e:
        raise Exception(error_msg.format(url_zip, e.errno, e.reason))
    except (Exception, KeyboardInterrupt) as a:
      if os.path.exists(fpath):
        os.remove(fpath)
      raise

def extract_archive(file_path, path='.', archive_format='auto'):
  """
  from:
      https://github.com/tensorflow/tensorflow/blob/v2.3.1/tensorflow/python/keras/utils/data_utils.py#L168-L297
      
  Extracts an archive if it matches tar, tar.gz, tar.bz, or zip formats.
  Arguments:
      file_path: path to the archive file
      path: path to extract the archive file
      archive_format: Archive format to try for extracting the file.
          Options are 'auto', 'tar', 'zip', and None.
          'tar' includes tar, tar.gz, and tar.bz files.
          The default 'auto' is ['tar', 'zip'].
          None or an empty list will return no matches found.
  Returns:
      True if a match was found and an archive extraction was completed,
      False otherwise.
  """
  if archive_format is None:
    return False
  if archive_format == 'auto':
    archive_format = ['tar', 'zip']
  if isinstance(archive_format, six.string_types):
    archive_format = [archive_format]

  #file_path = path_to_string(file_path)
  #path = path_to_string(path)

  for archive_type in archive_format:
    if archive_type == 'tar':
      open_fn = tarfile.open
      is_match_fn = tarfile.is_tarfile
    if archive_type == 'zip':
      open_fn = zipfile.ZipFile
      is_match_fn = zipfile.is_zipfile

    if is_match_fn(file_path):
      with open_fn(file_path) as archive:
        try:
          archive.extractall(path)
        except (tarfile.TarError, RuntimeError, KeyboardInterrupt):
          if os.path.exists(path):
            if os.path.isfile(path):
              os.remove(path)
            else:
              shutil.rmtree(path)
          raise
      return True
  return False

def updates_ondevice():
    """
    Searches the directory of AIDeveloper for available updates (zip folders)
    Returns
    list: list contains strings, each is a tag of an -update version    
    """
    files = os.listdir(dir_root)
    files = [file for file in files if file.startswith("AIDeveloper_") and file.endswith(".zip")]
    update_zips = [file for file in files if "update" in file]
    backup_zips = [file for file in files if "backup" in file]
    ondevice_zips = update_zips+backup_zips
    tags_update = [file.split("AIDeveloper_")[1] for file in ondevice_zips]
    tags_update = [file.split(".zip")[0] for file in tags_update]
    return tags_update

def check_for_updates(this_version):
    """
    check GitHub for new releases (updates of AIDeveloper)
    check local (on device) AIDeveloper folder for "-update.zip" and "-backup.zip" files
    
    Parameters
    ----------
    this_version: str, version of the installed release
    """
    #pre-define some variables
    Errors = None
    latest_release = None
    url = ""
    changelog = ""
    tags_update_online = []
    tags_update_ondevice = []

    #check local (on device)
    tags_update_ondevice = updates_ondevice()
    
    #check online
    try:
        url_releases = "https://github.com/maikherbig/AIDeveloper/releases"
        content = urllib.request.urlopen(url_releases).read().decode('UTF-8')
        tags = content.split("/maikherbig/AIDeveloper/releases/tag/")[1:]
        tags = [t.split('">AIDeveloper')[0] for t in tags]

        #handle  -update versions separately
        tags_update_online = [a for a in tags if "update" in a]
        tags_update_online = [a.split("-update")[0]+"-update" for a in tags_update_online]+["Bleeding edge"]
        tags = [a for a in tags if not "update" in a]

        latest_release = tags[0]
        highest_dev = [latest_release+"_dev" in tag for tag in tags]
        ind = np.where(np.array(highest_dev)==True)[0]
        if len(ind)>0:
            highest_devs_tag = list(np.array(tags)[ind])
            #corresponding numbers
            devs_tag_values =  [list(map(int, re.findall(r'\d+', t))) for t in highest_devs_tag]
            highest_dev_ind = np.argmax([t[3] for t in devs_tag_values])
            #name of the highest dev
            latest_release = highest_devs_tag[highest_dev_ind]
            url = "https://github.com/maikherbig/AIDeveloper/releases/tag/"+latest_release
        else:
            url = "https://github.com/maikherbig/AIDeveloper/releases/tag/"+latest_release
        
        #if "latest_release" is different from "this_version" then, an update is available!
        if this_version==latest_release:
            #no need to update, overwrite latest_release variable
            latest_release = "You are up to date"
            changelog = "You are up to date"
        else:
            #Open the url of the latest release and get the changelog
            content = urllib.request.urlopen(url).read().decode('UTF-8')
            changelog = content.split('<div class="markdown-body">')[1].split("</p>\n  </div>")[0]
            changelog = "Changelog:\n"+changelog.lstrip()
        
    except Exception as e:
        #There is an issue. Maybe no internet connection...
        Errors = e
        
    dic = {"Errors":Errors,"latest_release":latest_release,"latest_release_url":url
           ,"changelog":changelog,"tags_update_online":tags_update_online,
           "tags_update_ondevice":tags_update_ondevice}

    return dic

def aideveloper_filelist():
    files = [
    "AIDeveloper.py",
    "aid_backbone.py",
    "aid_bin.py",
    "aid_dependencies_linux.txt",
    "aid_dependencies_mac.txt",
    "aid_dependencies_win.txt",
    "aid_dl.py",
    "aid_frontend.py",
    "aid_img.py",
    "aid_imports.py",
    "aid_start.py",
    "aid_settings.json",
    "Empty.rtdc",
    "layout_dark_notooltip.txt",
    "layout_dark.txt",
    "layout_darkorange_notooltip.txt",
    "layout_darkorange.txt",
    "main_icon_simple_04_48.ico",
    "main_icon_simple_04_256.icns",
    "main_icon_simple_04_256.ico",
    "model_zoo.py",
    "partial_trainability.py"]
    return files

def check_aid_scripts_complete(aid_directory):
    files = aideveloper_filelist()
    #complete path to files/folders
    files_paths = [os.path.join(aid_directory,file) for file in files] #full path

    #check that all required files are there    
    check = [os.path.isfile(path) or os.path.isdir(path) for path in files_paths]
    ind = [i for i, x in enumerate(check) if not x]       
    assert all(check), "Cannot create a backup! Following file(s) are missing: "+str([files_paths[i] for i in ind])

    return files

    
def backup_current_version(VERSION):
    """
    Collect following files and folder:
    """
    files = check_aid_scripts_complete(dir_root)

    #create a name for the zipfile (without overwriting) 
    path_save = "AIDeveloper_"+VERSION+"-backup.zip"
    path_save = os.path.join(dir_root,path_save)#path to save the update zip        
    if not os.path.exists(path_save):#if such a file does not yet exist...
        path_save = path_save
    else:#such a file already exists!
        #Avoid to overwriting existing file:
        print("Adding additional number since file exists!")
        i = 1
        while os.path.exists(path_save):
            path_save = "AIDeveloper_"+VERSION+"-backup_"+str(i)+".zip"
            path_save = os.path.join(dir_root,path_save)#path to save the update zip        
            i+=1

    #create a ZipFile object
    with zipfile.ZipFile(path_save, 'w',compression=zipfile.ZIP_DEFLATED) as zipObj:
        for file in files:
            path_orig = os.path.join(dir_root,file)
            zipObj.write(path_orig, os.path.basename(path_orig))

        dirName = os.path.join(dir_root,"art")
        #Iterate over all the files in "art" folder
        for folder, subfolders, filenames in os.walk(dirName):
            for filename in filenames:
                filePath = os.path.join(folder, filename)
                path_in_zip = os.path.join("art",filePath.split(os.sep+"art"+os.sep)[1])
                zipObj.write(filePath,path_in_zip)
               
    return path_save #return the filename of the backup file

def delete_current_version(delete_art=False):
    files = check_aid_scripts_complete(dir_root)
    files_paths = [os.path.join(dir_root,file) for file in files] #full path
    for file in files_paths:
        if os.path.isfile(file):#if file exists
            os.remove(file)
    if delete_art:
        try:
            file = os.path.join(dir_root,"art")
            shutil.rmtree(file)
        except:
            pass
   
def download_aid_update(tag):
    url_zip = "https://github.com/maikherbig/AIDeveloper/releases/download/"+tag+"/AIDeveloper_"+tag+".zip"
    path_save = "AIDeveloper_"+tag+".zip"
    path_save = os.path.join(dir_root,path_save)#path to save the update zip
    if not os.path.isfile(path_save):
        download_zip(url_zip,path_save)
        return {"success":True,"path_save":path_save}
    else:
        return {"success":False,"path_save":path_save}

def download_aid_repo():
    #Check online for most recent scripts (bleeding edge update)
    files = aideveloper_filelist()#+["art"]#TODO: also download the folder art
    url_scripts_repo = ["https://raw.github.com/maikherbig/AIDeveloper/master/AIDeveloper/"+f for f in files]
    path_temp = create_temp_folder()
    path_temp = os.path.join(path_temp,"update_files_bleeding_edge")
    
    try:
        if os.path.exists(path_temp):
            #delete this folder
            shutil.rmtree(path_temp)
        #Create this folder (nice and empty)
        os.mkdir(path_temp)
    
        for i  in range(len(files)):
            save_to = os.path.join(path_temp,files[i])
            download_zip(url_scripts_repo[i],save_to)
        
        #check completeness/integrity (certain scripts need to be there)
        check_aid_scripts_complete(path_temp)
        
        #create a zip file 
        date = datetime.datetime.now().strftime("%Y%m%d_%H-%M-%S")
        path_save = os.path.join(dir_root,"AIDeveloper_"+date)
        shutil.make_archive(path_save, 'zip', path_temp)
        #delete the temporary files
        shutil.rmtree(path_temp)
        return {"success":True,"path_save":path_save+".zip"}
    except:
        return {"success":False,"path_save":""}

def update_from_zip(item_path,VERSION):
    #check that the zip contains all required files
    #zip_content = zipfile.ZipFile(item_path, 'r').namelist()
    
    #Unzip it into temp/update_files (only to check that everything is complete)
    #these files will actally be deleted immediately after checking
    path_temp = create_temp_folder()
    path_temp = os.path.join(path_temp,"update_files")
    extract_archive(item_path, path_temp, archive_format='zip')
    #check completeness/integrity (certain scripts need to be there)
    check_aid_scripts_complete(path_temp)
    #delete these files again
    shutil.rmtree(path_temp)

    #create a backup of the current version
    path_backup = backup_current_version(VERSION)
    #delete current version
    if item_path.endswith("update.zip"):#A release was downloaded. There, an art folder is contained
        #Delete all scripts and the art folder
        delete_current_version(delete_art=True)
    else:
        #Delete all scripts but keep the art folder
        delete_current_version(delete_art=False)
        
    #Unzip update into dir_root
    extract_archive(item_path, dir_root, archive_format='zip')

    #check completeness/integrity again (certain scripts need to be there)
    try:
        check_aid_scripts_complete(dir_root)
    except:#restore previous version from backup.zip 
        extract_archive(path_backup, dir_root, archive_format='zip')
    return path_backup

def bin_num_doane(a):
    """Compute number of bins based on Doane's formula
    Notes
    -----
    If the bin width cannot be determined, then a bin
    number of 5 is returned.
    See Also
    --------
    bin_width_doane: method used to compute the bin width
    """
    bad = np.isnan(a) | np.isinf(a)
    data = a[~bad]
    acc = bin_width_doane(a)
    if acc == 0 or np.isnan(acc):
        num = 5
    else:
        num = int(np.round((data.max() - data.min()) / acc))
    return num


def bin_width_doane(a):
    """Get a sensible bin width for a (2D) histogram, based on Doanes rule
    References
    ----------
    - `<https://en.wikipedia.org/wiki/Histogram#Number_of_bins_and_width>`_
    - `<https://stats.stackexchange.com/questions/55134/
      doanes-formula-for-histogram-binning>`_
    -----
    """
    bad = np.isnan(a) | np.isinf(a)
    data = a[~bad]
    n = data.size
    g1 = skew(data)
    sigma_g1 = np.sqrt(6 * (n - 2) / ((n + 1) * (n + 3)))
    k = 1 + np.log2(n) + np.log2(1 + np.abs(g1) / sigma_g1)
    acc = (data.max() - data.min()) / k
    return acc


def kde_histogram(x, y, xout=None, yout=None, bins=None):
    """ Histogram-based Kernel Density Estimation
    Parameters
    ----------
    x, y: 1D array
        Input values for kernel density estimation.
    xout, yout: 1D arrays
        Coordinates at which the density should be determined. If set to none
        the input values are used.
    bins: tuple (binsx, binsy)
        The number of bins in x and y direction to construct the histogram.
    Returns
    -------
    density: 1D array
        The density estimatation for the points (xout, yout)
    See Also
    --------
    `numpy.histogram2d`
    `scipy.interpolate.RectBivariateSpline`
    """
    valid_combination = ((xout==yout==None) or
                   (xout is not None and yout is not None))
    if not valid_combination:
        raise ValueError("xout and yout must either be None, or provide values")
    
    if xout is not None and yout is not None:
        valid_length = len(xout)==len(yout)
        if not valid_length:
            raise ValueError("xout and yout must have the same length")

    else:#if xout==yout==None:
        xout = x
        yout = y

    if bins is None:
        bins = (max(5, bin_num_doane(x)),max(5, bin_num_doane(y)))

    # Compute the histogram
    hist2d, xedges, yedges = np.histogram2d(x=x,y=y,bins=bins,normed=True)
    xip = xedges[1:]-(xedges[1]-xedges[0])/2
    yip = yedges[1:]-(yedges[1]-yedges[0])/2

    estimator = RectBivariateSpline(x=xip, y=yip, z=hist2d)
    density = estimator.ev(xout, yout)
    density[density < 0] = 0
    
    density = density.reshape(xout.shape)
    density = density-np.min(density)
    density = density/np.max(density)
    return density


def kde_gauss(y, x, xout=None, yout=None):
    """ Gaussian Kernel Density Estimation
    Parameters
    ----------
    x, y: 1D array
        Input values for kernel density estimation.
    xout, yout: 1D arrays
        Coordinates at which the density is determined.
        If set to none, input values are used.
    Returns
    -------
    density: 1D array

    See Also
    --------
    `scipy.stats.gaussian_kde`
    """
    valid_combination = ((xout==yout==None) or
                   (xout is not None and yout is not None))
    if not valid_combination:
        raise ValueError("xout and yout must either be None, or provide values")
    
    if xout is not None and yout is not None:
        valid_length = len(xout)==len(yout)
        if not valid_length:
            raise ValueError("xout and yout must have the same length")

    else:#if xout==yout==None:
        xout = x
        yout = y

    try:
        estimator = gaussian_kde([x.flatten(), y.flatten()])
        density = estimator.evaluate([xout.flatten(), yout.flatten()])
    except np.linalg.LinAlgError:
        # Error occurs when just a single values is provided
        density = np.empty(xout.shape)*np.nan
    density = density.reshape(xout.shape)
    density = density-np.min(density)
    density = density/np.max(density)
    return density









#import aid_start #import a module that sits in the AIDeveloper folder
#dir_root = os.path.dirname(aid_start.__file__)#ask the module for its origin

#def save_aid_settings(Default_dict):
#    dir_settings = os.path.join(dir_root,"aid_settings.json")#dir to settings
#    #Save the layout to Default_dict
#    with open(dir_settings, 'w') as f:
#        json.dump(Default_dict,f)



  

#################Some functions that are not used anymore######################

#def updates_online():
#    """
#    Searches the directory of AIDeveloper for available updates (zip folders)
#    Returns
#    list: list contains strings, each is a tag of an "a.b.c-update" version    
#    """
#    url_releases = "https://github.com/maikherbig/AIDeveloper/releases"
#    content = urllib.request.urlopen(url_releases).read().decode('UTF-8')
#    tags = content.split("/maikherbig/AIDeveloper/releases/tag/")[1:]
#    tags = [t.split('">AIDeveloper')[0] for t in tags]
#    tags_update = [a for a in tags if "update" in a]
#    tags_update = [a.split("-update")[0]+"-update" for a in tags_update]
#    return tags_update

#def zoom(a, factor):
#    a = np.asarray(a)
#    slices = [slice(0, old, 1/factor) for old in a.shape]
#    idxs = (np.mgrid[slices]).astype('i')
#    return a[tuple(idxs)]


#def load_model_json_h5(loc_json,loc_h5=False):
#    # load json and create model
#    json_file = open(loc_json, 'r')
#    loaded_model_json = json_file.read()
#    json_file.close()
#    loaded_model = model_from_json(loaded_model_json)
#    if loc_h5:
#        # load weights into new model
#        loaded_model.load_weights(loc_h5)
#    return loaded_model

#class MyThread(QtCore.QThread):
#    def run(self):
#        self.exec_()






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 24 13:13:59 2022

@author: nana
"""
import tensorflow as tf
import numpy as np
import pandas as pd
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Dense, Activation
from tensorflow.keras import backend as K

import keras_metrics
#from . import keras_metrics #side package for precision, recall etc during training
global keras_metrics


class cyclicLR(tf.keras.callbacks.Callback):
    """
    Reference: https://github.com/bckenstler/CLR/blob/master/clr_callback.py
    Author: Brad Kenstler
    
    This callback implements a cyclical learning rate policy (CLR).
    The method cycles the learning rate between two boundaries with
    some constant frequency, as detailed in this paper (https://arxiv.org/abs/1506.01186).
    The amplitude of the cycle can be scaled on a per-iteration or 
    per-cycle basis.
    This class has three built-in policies, as put forth in the paper.
    "triangular":
        A basic triangular cycle w/ no amplitude scaling.
    "triangular2":
        A basic triangular cycle that scales initial amplitude by half each cycle.
    "exp_range":
        A cycle that scales initial amplitude by gamma**(cycle iterations) at each 
        cycle iteration.
    For more detail, please see paper.
    
    # Example
        ```python
            clr = cyclicLR(base_lr=0.001, max_lr=0.006,
                                step_size=2000., mode='triangular')
            model.fit(X_train, Y_train, callbacks=[clr])
        ```
    
    Class also supports custom scaling functions:
        ```python
            clr_fn = lambda x: 0.5*(1+np.sin(x*np.pi/2.))
            clr = cyclicLR(base_lr=0.001, max_lr=0.006,
                                step_size=2000., scale_fn=clr_fn,
                                scale_mode='cycle')
            model.fit(X_train, Y_train, callbacks=[clr])
        ```    
    # Arguments
        base_lr: initial learning rate which is the
            lower boundary in the cycle.
        max_lr: upper boundary in the cycle. Functionally,
            it defines the cycle amplitude (max_lr - base_lr).
            The lr at any cycle is the sum of base_lr
            and some scaling of the amplitude; therefore 
            max_lr may not actually be reached depending on
            scaling function.
        step_size: number of training iterations per
            half cycle. Authors suggest setting step_size
            2-8 x training iterations in epoch.
        mode: one of {triangular, triangular2, exp_range}.
            Default 'triangular'.
            Values correspond to policies detailed above.
            If scale_fn is not None, this argument is ignored.
        gamma: constant in 'exp_range' scaling function:
            gamma**(cycle iterations)
        scale_fn: Custom scaling policy defined by a single
            argument lambda function, where 
            0 <= scale_fn(x) <= 1 for all x >= 0.
            mode paramater is ignored 
        scale_mode: {'cycle', 'iterations'}.
            Defines whether scale_fn is evaluated on 
            cycle number or cycle iterations (training
            iterations since start of cycle). Default is 'cycle'.
    """

    def __init__(self, base_lr=0.001, max_lr=0.006, step_size=2000., mode='triangular',
                 gamma=1., scale_fn=None, scale_mode='cycle'):
        super(cyclicLR, self).__init__()

        self.base_lr = base_lr
        self.max_lr = max_lr
        self.step_size = step_size
        self.mode = mode
        self.gamma = gamma
        if scale_fn == None:
            if self.mode == 'triangular':
                self.scale_fn = lambda x: 1.
                self.scale_mode = 'cycle'
            elif self.mode == 'triangular2':
                self.scale_fn = lambda x: 1/(2.**(x-1))
                self.scale_mode = 'cycle'
            elif self.mode == 'exp_range':
                self.scale_fn = lambda x: gamma**(x)
                self.scale_mode = 'iterations'
        else:
            self.scale_fn = scale_fn
            self.scale_mode = scale_mode
        self.clr_iterations = 0.
        self.trn_iterations = 0.
        self.history = {}

        self._reset()

    def _reset(self, new_base_lr=None, new_max_lr=None,
               new_step_size=None):
        """Resets cycle iterations.
        Optional boundary/step size adjustment.
        """
        if new_base_lr != None:
            self.base_lr = new_base_lr
        if new_max_lr != None:
            self.max_lr = new_max_lr
        if new_step_size != None:
            self.step_size = new_step_size
        self.clr_iterations = 0.
        
    def clr(self):
        cycle = np.floor(1+self.clr_iterations/(2*self.step_size))
        x = np.abs(self.clr_iterations/self.step_size - 2*cycle + 1)
        if self.scale_mode == 'cycle':
            return self.base_lr + (self.max_lr-self.base_lr)*np.maximum(0, (1-x))*self.scale_fn(cycle)
        else:
            return self.base_lr + (self.max_lr-self.base_lr)*np.maximum(0, (1-x))*self.scale_fn(self.clr_iterations)
        
    def on_train_begin(self, logs={}):
        logs = logs or {}

        if self.clr_iterations == 0:
            K.set_value(self.model.optimizer.lr, self.base_lr)
        else:
            K.set_value(self.model.optimizer.lr, self.clr())        
            
    def on_batch_end(self, epoch, logs=None):
        
        logs = logs or {}
        self.trn_iterations += 1
        self.clr_iterations += 1

        self.history.setdefault('lr', []).append(K.get_value(self.model.optimizer.lr))
        self.history.setdefault('iterations', []).append(self.trn_iterations)

        for k, v in logs.items():
            self.history.setdefault(k, []).append(v)
        
        K.set_value(self.model.optimizer.lr, self.clr())


class exponentialDecay(tf.keras.callbacks.Callback):
    """    
    Decrease learning rate using exponential decay
    lr = initial_lr * decay_rate ** (batch_iteration / decay_steps)    
    """
    def __init__(self, initial_lr=0.01, decay_steps=100, decay_rate=0.95):
        super(exponentialDecay, self).__init__()

        self.initial_lr = initial_lr
        self.decay_steps = decay_steps
        self.decay_rate = decay_rate

        self.iterations = 0.
        self.trn_iterations = 0.
        self.history = {}

        self._reset()

    def _reset(self, new_initial_lr=None, new_decay_steps=None,new_decay_rate=None):
        if new_initial_lr != None:
            self.initial_lr = new_initial_lr
        if new_decay_steps != None:
            self.decay_steps = new_decay_steps
        if new_decay_rate != None:
            self.decay_rate = new_decay_rate
        self.iterations = 0.
        
    def exp_decay(self):
        return self.initial_lr * self.decay_rate ** (self.iterations / self.decay_steps)

    def on_train_begin(self, logs={}):
        logs = logs or {}

        if self.iterations == 0:
            K.set_value(self.model.optimizer.lr, self.initial_lr)
        else:
            K.set_value(self.model.optimizer.lr, self.exp_decay())        
            
    def on_batch_end(self, epoch, logs=None):
        
        logs = logs or {}
        self.trn_iterations += 1
        self.iterations += 1

        self.history.setdefault('lr', []).append(K.get_value(self.model.optimizer.lr))
        self.history.setdefault('iterations', []).append(self.trn_iterations)

        for k, v in logs.items():
            self.history.setdefault(k, []).append(v)
        
        K.set_value(self.model.optimizer.lr, self.exp_decay())


def get_config(cpu_nr,gpu_nr,deviceSelected,gpu_memory):
    #No GPU available, CPU selected:
    if gpu_nr==0: #and deviceSelected=="Default CPU":
        print("Adjusted options for CPU usage")
        config_gpu = tf.compat.v1.ConfigProto()
    #GPU available but user wants to use CPU
    elif gpu_nr>0 and deviceSelected=="Default CPU":
        config_gpu = tf.compat.v1.ConfigProto(intra_op_parallelism_threads=cpu_nr,\
                inter_op_parallelism_threads=cpu_nr, allow_soft_placement=True,\
                device_count = {'CPU' : 1, 'GPU' : 0})
        print("Adjusted options for CPU usage")
    
    #GPU selected
    elif deviceSelected=="Single-GPU" or deviceSelected=="Multi-GPU":
        gpu_options = tf.compat.v1.GPUOptions(per_process_gpu_memory_fraction=gpu_memory,\
                                    allow_growth = True)
        config_gpu = tf.compat.v1.ConfigProto(allow_soft_placement=True,\
                                    gpu_options = gpu_options,\
                                    log_device_placement=False)
        if deviceSelected=="Single-GPU":
            print("Adjusted GPU options for Single-GPU usage. Set memeory fraction to "+str(gpu_memory))
        if deviceSelected=="Multi-GPU":
            print("Adjusted GPU options for Multi-GPU usage. Set memeory fraction to "+str(gpu_memory))
    

    return config_gpu

def model_in_out_dim(model_object,object_type):
    if object_type=="config":
    #user provided model_config:
        try: #Sequential Model
            in_dim = model_object['config'][0]['config']['batch_input_shape']
        except: #Functional Api
            in_dim = model_object['config']["layers"][0]["config"]["batch_input_shape"]
        try: #Sequential Model
            out_dim = model_object['config'][-2]['config']['units']
        except: #Functional Api
            out_dim = model_object['config']["layers"][-2]["config"]["units"]

    if object_type=="model":
        in_dim = model_object.layers[0].get_input_shape_at(0)
        out_dim = model_object.layers[-1].get_output_shape_at(0)[1]

    return in_dim,out_dim

def model_add_classes(model_keras,nr_classes):
    #Sequential or Functional API?
    model_config = model_keras.get_config()
    if "sequential" not in model_config["name"]:
        print("Loaded model is defined using functional API of Keras")
        model_api = "functional"
    if "sequential" in model_config["name"]:
        print("Loaded model is defined using the sequential API of Keras")
        model_api = "sequential"
    
    #get list of layer types
    layer_types_list = [l.__class__.__name__ for l in model_keras.layers]
    #get the index of the last dense layer
    ind = [name=="Dense" for name in layer_types_list]#where are dense layers
    ind = np.where(np.array(ind)==True)[0][-1]#index of the last dense layer
    last_dense_name = model_keras.layers[ind].name #name of that last dense layer
    ind = ind-1 #go even one more layer back, because we need its output
    
    #For functional API:
    if model_api=="functional":
        #overwrite model_keras with a new, (shortened) model
        model_keras = Model(model_keras.get_input_at(0),model_keras.layers[ind].output)
        #Add a new final dense layer
        x = Dense(nr_classes,name=last_dense_name)(model_keras.layers[-1].output)
        x = Activation('softmax',name="outputTensor")(x)
        model_keras = Model(inputs=model_keras.get_input_at(0), outputs=x)
    #For sequential API:
    elif model_api=="sequential":
        model_keras.pop()#remove final activation layer
        model_keras.pop()#remove last dense layer
        model_keras.add(Dense(nr_classes,name=last_dense_name))
        model_keras.add(Activation('softmax',name="outputTensor"))
    
    #Compile to reset the optimizer weights
    model_keras.compile(optimizer='adam', loss='categorical_crossentropy')

def get_metrics_tensors(metrics,nr_classes):
    """
    Function takes a list of metrics and creates fresh tensors accordingly.
    This is necessary, after re-compiling the model because the placeholder has
    to be updated    
    """
    f1 = any(["auc" in str(a) for a in metrics])
    precision = any(["precision" in str(a) for a in metrics])
    recall = any(["recall" in str(a) for a in metrics])
    
    Metrics =  []
    if f1==True:
        for class_ in range(nr_classes):
            Metrics.append(tf.keras.metrics.AUC())
    if precision==True:
        for class_ in range(nr_classes):
            Metrics.append(tf.keras.metrics.Precision())
    if recall==True:
        for class_ in range(nr_classes):
            Metrics.append(tf.keras.metrics.Recall())
    metrics =  ['accuracy'] + Metrics
    return metrics

def model_compile(model_keras,loss_,optimizer_,learning_rate_,model_metrics,out_dim):
    print("Compiling")
    optimizer_name = optimizer_["comboBox_optimizer"].lower()
    print("optimizer_name:",optimizer_name)
    if optimizer_name=='sgd':
        optimizer = tf.keras.optimizers.SGD(learning_rate=optimizer_["doubleSpinBox_lr_sgd"], momentum=optimizer_["doubleSpinBox_sgd_momentum"], nesterov=optimizer_["checkBox_sgd_nesterov"])
    elif optimizer_name=='rmsprop':
        optimizer = tf.keras.optimizers.RMSprop(learning_rate=optimizer_["doubleSpinBox_lr_rmsprop"], rho=optimizer_["doubleSpinBox_rms_rho"])
    elif optimizer_name=='adagrad':
        optimizer = tf.keras.optimizers.Adagrad(learning_rate=optimizer_["doubleSpinBox_lr_adagrad"])
    elif optimizer_name=='adadelta':
        optimizer = tf.keras.optimizers.Adadelta(learning_rate=optimizer_["doubleSpinBox_lr_adadelta"], rho=optimizer_["doubleSpinBox_adadelta_rho"])
    elif optimizer_name=='adam':
        optimizer = tf.keras.optimizers.Adam(learning_rate=optimizer_["doubleSpinBox_lr_adam"], beta_1=optimizer_["doubleSpinBox_adam_beta1"], beta_2=optimizer_["doubleSpinBox_adam_beta2"], amsgrad=optimizer_["checkBox_adam_amsgrad"])
    elif optimizer_name=='adamax':
        optimizer = tf.keras.optimizers.Adamax(learning_rate=optimizer_["doubleSpinBox_lr_adamax"], beta_1=optimizer_["doubleSpinBox_adamax_beta1"], beta_2=optimizer_["doubleSpinBox_adamax_beta2"])
    elif optimizer_name=='nadam':
        optimizer = tf.keras.optimizers.Nadam(learning_rate=optimizer_["doubleSpinBox_lr_nadam"], beta_1=optimizer_["doubleSpinBox_nadam_beta1"], beta_2=optimizer_["doubleSpinBox_nadam_beta2"])
    else:
        print("Unknown optimizer!")
    model_keras.compile(loss=loss_,optimizer=optimizer,
                        metrics=get_metrics_tensors(model_metrics,out_dim))
    print("Compiling DONE")

    return 1

def model_get_trainable_list(model_keras):
    params = [model_keras.layers[i].count_params() for i in range(len(model_keras.layers))]
    #Only count layers that have parameters
    l_para_not_zero = np.where(np.array(params)>0)[0] #indices of layers that have parameters
    #Original trainable states of layers with parameters
    trainable_list = [model_keras.layers[i].trainable for i in l_para_not_zero]
    layer_names_list = [model_keras.layers[i].__class__.__name__  for i in l_para_not_zero]
    return trainable_list,layer_names_list

def get_dropout(model_keras):
    #Funktion takes a keras model and gives a list of the dropout values.    
    l_names = [model_keras.layers[i].__class__.__name__ for i in range(len(model_keras.layers))]
    ind = ["Dropout" in x for x in l_names]
    ind = np.where(np.array(ind)==True)[0] #at which indices are dropout layers?
    do_list = [model_keras.layers[i].rate for i in ind]
    return do_list

def get_optimizer_name(model_keras):
    optimizer = model_keras.optimizer.get_config()["name"]
    # optimizer = str(model_keras.optimizer)
    # optimizer = optimizer.split("<keras.optimizer_v2.")[1].split(" object at")[0]
    # optimizer = optimizer.split(".")[1]
    if optimizer in ['SGD','RMSprop','Adagrad',"Adamax",'Adadelta','Adam','Nadam']:
        return optimizer
    else:
        return 0

def get_optimizer_settings():
    """
    Default optimizer settings (taken from https://keras.io/api/optimizers)
    """
    d = {}
    d["comboBox_optimizer"] = "Adam"

    d["doubleSpinBox_lr_sgd"] = 0.01
    d["doubleSpinBox_sgd_momentum"] = 0.0
    d["checkBox_sgd_nesterov"] = False

    d["doubleSpinBox_lr_rmsprop"] = 0.001
    d["doubleSpinBox_rms_rho"] = 0.9

    d["doubleSpinBox_lr_adam"] = 0.001
    d["doubleSpinBox_adam_beta1"] = 0.9
    d["doubleSpinBox_adam_beta2"] = 0.999
    d["checkBox_adam_amsgrad"] = False

    d["doubleSpinBox_lr_nadam"] = 0.002
    d["doubleSpinBox_nadam_beta1"] = 0.9
    d["doubleSpinBox_nadam_beta2"] = 0.999

    d["doubleSpinBox_lr_adadelta"] = 1.0
    d["doubleSpinBox_adadelta_rho"] = 0.95

    d["doubleSpinBox_lr_adagrad"] = 0.01

    d["doubleSpinBox_lr_adamax"] = 0.002
    d["doubleSpinBox_adamax_beta1"] = 0.9
    d["doubleSpinBox_adamax_beta2"] = 0.999

    return d

def get_cyclStepSize(SelectedFiles,step_size,batch_size):
    ind = [selectedfile["TrainOrValid"] == "Train" for selectedfile in SelectedFiles]
    ind = np.where(np.array(ind)==True)[0]
    SelectedFiles_train = np.array(SelectedFiles)[ind]
    SelectedFiles_train = list(SelectedFiles_train)
    nr_events_train_total = np.sum([int(selectedfile["nr_events_epoch"]) for selectedfile in SelectedFiles_train])

    cycLrStepSize = step_size*int(np.round(nr_events_train_total / batch_size))#number of steps in one epoch 
    return cycLrStepSize      

def model_change_trainability(model_keras,trainable_new,model_metrics,out_dim,loss_,optimizer_,learning_rate_):
    print("Line 83")
    #Function takes a keras model and list of trainable states.
    #The last n layers, which have parameters (not activation etc.) are set to 'not trainable'
    params = [model_keras.layers[i].count_params() for i in range(len(model_keras.layers))]
    #Only count layers that have parameters
    l_para_not_zero = np.where(np.array(params)>0)[0] #indices of layers that have parameters
    nr_layers = len(l_para_not_zero) #total nr. of layers of model with parameters
    #check that the number of layers in model is equal to the len(trainable_new) 
    assert nr_layers==len(trainable_new)
    #Original trainable states of layers with parameters
    trainable_now = [model_keras.layers[i].trainable for i in l_para_not_zero]
    if trainable_now == trainable_new:#they are already equal
        text = "No need to change trainable states of model."
    else:
        print("Change trainable states of model.")
        #Now change the trainability of each layer with parameters to the state given in trainable_states
        for i in range(nr_layers):
            layer_index = l_para_not_zero[i]
            model_keras.layers[layer_index].trainable = trainable_new[i]   
        #Model has to be recompiled in order to see any any effect of the change
        model_compile(model_keras,loss_,optimizer_,learning_rate_,model_metrics,out_dim)
    
        text = []
        model_keras.summary(print_fn=text.append)
        text = "\n".join(text)
    return text

#Define some custom metrics which will allow to use precision, recall, etc during training
def get_custom_metrics():
    custom_metrics = []
    #Keep keras_metrics to maintain compatibiliy with AID_0.2.x versions
    custom_metrics.append(keras_metrics.categorical_f1_score())
    custom_metrics.append(keras_metrics.categorical_precision())
    custom_metrics.append(keras_metrics.categorical_recall())
    custom_metrics = {m.__name__: m for m in custom_metrics}
    custom_metrics["sin"] = K.sin
    custom_metrics["abs"] = K.abs
    return custom_metrics

def get_lr_dict(learning_rate_const_on,learning_rate_const,
                    learning_rate_cycLR_on,cycLrMin,cycLrMax,
                    cycLrMethod,cycLrStepSize,
                    learning_rate_expo_on,
                    expDecInitLr,expDecSteps,expDecRate,cycLrGamma):
    lr_dict = pd.DataFrame()
    lr_dict["learning_rate_const_on"] = learning_rate_const_on,
    lr_dict["learning_rate_const"] = learning_rate_const,
    lr_dict["learning_rate_cycLR_on"] = learning_rate_cycLR_on,
    lr_dict["cycLrMin"] = cycLrMin,
    lr_dict["cycLrMax"] = cycLrMax,
    lr_dict["cycLrMethod"] = cycLrMethod,
    lr_dict["cycLrStepSize"] = cycLrStepSize,
    lr_dict["expDecInitLr"] = expDecInitLr,
    lr_dict["expDecInitLr"] = expDecInitLr,
    lr_dict["expDecSteps"] = expDecSteps,
    lr_dict["expDecRate"] = expDecRate,
    lr_dict["cycLrGamma"] = cycLrGamma,
    return lr_dict

def get_lr_callback(learning_rate_const_on,learning_rate_const,
                    learning_rate_cycLR_on,cycLrMin,cycLrMax,
                    cycLrMethod,cycLrStepSize,
                    learning_rate_expo_on,
                    expDecInitLr,expDecSteps,expDecRate,cycLrGamma):

    if learning_rate_const_on==True:
        return None
    elif learning_rate_cycLR_on==True:
        return cyclicLR(base_lr=cycLrMin,max_lr=cycLrMax,step_size=cycLrStepSize,mode=cycLrMethod,gamma=cycLrGamma)  
    elif learning_rate_expo_on==True:
        return exponentialDecay(initial_lr=expDecInitLr, decay_steps=expDecSteps, decay_rate=expDecRate)  
    
def change_dropout(model_keras,do,model_metrics,out_dim,loss_,optimizer_,learning_rate_):
    #Funktion takes a keras model and changes the dropout.
    #do = float (0 to 1) or list (with values from 0 to 1)
    #if do is a list: len(do) has to be equal to the nr. of dropout layers in the model
    #if do is a float: this value is then used for all droput layers
    model_keras._make_train_function()   
    model_weights = model_keras.get_weights() # Store weights
    optimizer_weights = model_keras.optimizer.get_weights() # Store optimizer values
    
    l_names = [model_keras.layers[i].__class__.__name__ for i in range(len(model_keras.layers))]
    ind = ["Dropout" in x for x in l_names]
    ind = np.where(np.array(ind)==True)[0] #at which indices are dropout layers?
    if len(ind)==0:#if there are no dropout layers, return
        return 0
    if type(do)==float:
        #set all dropout layers to do
        for i in range(len(ind)):
            model_keras.layers[ind[i]].rate = do
    elif type(do)==list:
        #check that the length of the list equals the nr. of dropout layers
        if len(do)!=len(ind):
            return 0
        else:
            for i in range(len(ind)):
                model_keras.layers[ind[i]].rate = do[i]
                
    #Only way that it actually has an effect is to clone the model, compile and reload weights
    model_keras = tf.keras.models.clone_model(model_keras) # If I do not clone, the new rate is never used. Weights are re-init now.
    model_compile(model_keras,loss_,optimizer_,learning_rate_,model_metrics,out_dim)
    
    model_keras._make_train_function()
    model_keras.set_weights(model_weights) # Set weights
    model_keras.optimizer.set_weights(optimizer_weights)#Set optimizer values
    return 1


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import setuptools 

setuptools.setup(
      packages=setuptools.find_namespace_packages(
                     include=["napari_aideveloper", "napari_aideveloper.*"], ),
                 include_package_data=True
                 )



__version__ = "0.0.1"


from ._dock_widget import napari_experimental_provide_dock_widget
from .aid_bin import*
from .aid_dl import*
from . import aid_img
from .aid_start import*
from . import model_zoo





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 25 13:54:23 2022

@author: nana
"""

from inspect import getmembers, isfunction
import aid_bin, aid_dl, aid_start,aid_img, aid_backbone



functions_aid_bin = [o[0] for o in getmembers(aid_bin) if isfunction(o[1])]
functions_aid_dl = [o[0] for o in getmembers(aid_dl) if isfunction(o[1])]
functions_aid_start = [o[0] for o in getmembers(aid_start) if isfunction(o[1])]
functions_aid_img = [o[0] for o in getmembers(aid_img) if isfunction(o[1])]
functions_aid_backbone = [o[0] for o in getmembers(aid_backbone) if isfunction(o[1])]



functions_name = functions_aid_backbone
function_list = ""
for i in functions_name:
    function_list += i + ", "
    
print(function_list)
    